#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Union, Tuple, Optional

from pyspark.sql.streaming.stateful_processor_api_client import StatefulProcessorApiClient
from pyspark.sql.types import StructType
from pyspark.errors import PySparkRuntimeError

__all__ = ["ValueStateClient"]


class ValueStateClient:
    def __init__(
        self,
        stateful_processor_api_client: StatefulProcessorApiClient,
        schema: Union[StructType, str],
    ) -> None:
        self._stateful_processor_api_client = stateful_processor_api_client
        if isinstance(schema, str):
            self.schema = self._stateful_processor_api_client._parse_string_schema(schema)
        else:
            self.schema = schema

    def exists(self, state_name: str) -> bool:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        exists_call = stateMessage.Exists()
        value_state_call = stateMessage.ValueStateCall(stateName=state_name, exists=exists_call)
        state_variable_request = stateMessage.StateVariableRequest(valueStateCall=value_state_call)
        message = stateMessage.StateRequest(stateVariableRequest=state_variable_request)

        self._stateful_processor_api_client._send_proto_message(message.SerializeToString())
        response_message = self._stateful_processor_api_client._receive_proto_message()
        status = response_message[0]
        if status == 0:
            return True
        elif status == 2:
            # Expect status code is 2 when state variable doesn't have a value.
            return False
        else:
            # TODO(SPARK-49233): Classify user facing errors.
            raise PySparkRuntimeError(f"Error checking value state exists: {response_message[1]}")

    def get(self, state_name: str) -> Optional[Tuple]:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        get_call = stateMessage.Get()
        value_state_call = stateMessage.ValueStateCall(stateName=state_name, get=get_call)
        state_variable_request = stateMessage.StateVariableRequest(valueStateCall=value_state_call)
        message = stateMessage.StateRequest(stateVariableRequest=state_variable_request)

        self._stateful_processor_api_client._send_proto_message(message.SerializeToString())
        response_message = self._stateful_processor_api_client._receive_proto_message()
        status = response_message[0]
        if status == 0:
            if len(response_message[2]) == 0:
                return None
            data = self._stateful_processor_api_client._deserialize_from_bytes(response_message[2])
            return tuple(data)
        else:
            # TODO(SPARK-49233): Classify user facing errors.
            raise PySparkRuntimeError(f"Error getting value state: {response_message[1]}")

    def update(self, state_name: str, value: Tuple) -> None:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        bytes = self._stateful_processor_api_client._serialize_to_bytes(self.schema, value)
        update_call = stateMessage.ValueStateUpdate(value=bytes)
        value_state_call = stateMessage.ValueStateCall(
            stateName=state_name, valueStateUpdate=update_call
        )
        state_variable_request = stateMessage.StateVariableRequest(valueStateCall=value_state_call)
        message = stateMessage.StateRequest(stateVariableRequest=state_variable_request)

        self._stateful_processor_api_client._send_proto_message(message.SerializeToString())
        response_message = self._stateful_processor_api_client._receive_proto_message()
        status = response_message[0]
        if status != 0:
            # TODO(SPARK-49233): Classify user facing errors.
            raise PySparkRuntimeError(f"Error updating value state: {response_message[1]}")

    def clear(self, state_name: str) -> None:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        clear_call = stateMessage.Clear()
        value_state_call = stateMessage.ValueStateCall(stateName=state_name, clear=clear_call)
        state_variable_request = stateMessage.StateVariableRequest(valueStateCall=value_state_call)
        message = stateMessage.StateRequest(stateVariableRequest=state_variable_request)

        self._stateful_processor_api_client._send_proto_message(message.SerializeToString())
        response_message = self._stateful_processor_api_client._receive_proto_message()
        status = response_message[0]
        if status != 0:
            # TODO(SPARK-49233): Classify user facing errors.
            raise PySparkRuntimeError(f"Error clearing value state: {response_message[1]}")



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file

Licensed to the Apache Software Foundation (ASF) under one or more
contributor license agreements.  See the NOTICE file distributed with
this work for additional information regarding copyright ownership.
The ASF licenses this file to You under the Apache License, Version 2.0
(the "License"); you may not use this file except in compliance with
the License.  You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import builtins
import collections.abc
import google.protobuf.any_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.timestamp_pb2
import pyspark.sql.connect.proto.common_pb2
import pyspark.sql.connect.proto.expressions_pb2
import pyspark.sql.connect.proto.relations_pb2
import pyspark.sql.connect.proto.types_pb2
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _OutputType:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _OutputTypeEnumTypeWrapper(
    google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_OutputType.ValueType],
    builtins.type,
):  # noqa: F821
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    OUTPUT_TYPE_UNSPECIFIED: _OutputType.ValueType  # 0
    """Safe default value. Should not be used."""
    MATERIALIZED_VIEW: _OutputType.ValueType  # 1
    """A materialized view which is published to the catalog"""
    TABLE: _OutputType.ValueType  # 2
    """A table which is published to the catalog"""
    TEMPORARY_VIEW: _OutputType.ValueType  # 3
    """A view which is not published to the catalog"""
    SINK: _OutputType.ValueType  # 4
    """A sink which is not published to the catalog"""

class OutputType(_OutputType, metaclass=_OutputTypeEnumTypeWrapper):
    """The type of output."""

OUTPUT_TYPE_UNSPECIFIED: OutputType.ValueType  # 0
"""Safe default value. Should not be used."""
MATERIALIZED_VIEW: OutputType.ValueType  # 1
"""A materialized view which is published to the catalog"""
TABLE: OutputType.ValueType  # 2
"""A table which is published to the catalog"""
TEMPORARY_VIEW: OutputType.ValueType  # 3
"""A view which is not published to the catalog"""
SINK: OutputType.ValueType  # 4
"""A sink which is not published to the catalog"""
global___OutputType = OutputType

class PipelineCommand(google.protobuf.message.Message):
    """Dispatch object for pipelines commands. See each individual command for documentation."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class CreateDataflowGraph(google.protobuf.message.Message):
        """Request to create a new dataflow graph."""

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        class SqlConfEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: builtins.str
            value: builtins.str
            def __init__(
                self,
                *,
                key: builtins.str = ...,
                value: builtins.str = ...,
            ) -> None: ...
            def ClearField(
                self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]
            ) -> None: ...

        DEFAULT_CATALOG_FIELD_NUMBER: builtins.int
        DEFAULT_DATABASE_FIELD_NUMBER: builtins.int
        SQL_CONF_FIELD_NUMBER: builtins.int
        default_catalog: builtins.str
        """The default catalog."""
        default_database: builtins.str
        """The default database."""
        @property
        def sql_conf(
            self,
        ) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
            """SQL configurations for all flows in this graph."""
        def __init__(
            self,
            *,
            default_catalog: builtins.str | None = ...,
            default_database: builtins.str | None = ...,
            sql_conf: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
        ) -> None: ...
        def HasField(
            self,
            field_name: typing_extensions.Literal[
                "_default_catalog",
                b"_default_catalog",
                "_default_database",
                b"_default_database",
                "default_catalog",
                b"default_catalog",
                "default_database",
                b"default_database",
            ],
        ) -> builtins.bool: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal[
                "_default_catalog",
                b"_default_catalog",
                "_default_database",
                b"_default_database",
                "default_catalog",
                b"default_catalog",
                "default_database",
                b"default_database",
                "sql_conf",
                b"sql_conf",
            ],
        ) -> None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_default_catalog", b"_default_catalog"]
        ) -> typing_extensions.Literal["default_catalog"] | None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_default_database", b"_default_database"]
        ) -> typing_extensions.Literal["default_database"] | None: ...

    class DropDataflowGraph(google.protobuf.message.Message):
        """Drops the graph and stops any running attached flows."""

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        DATAFLOW_GRAPH_ID_FIELD_NUMBER: builtins.int
        dataflow_graph_id: builtins.str
        """The graph to drop."""
        def __init__(
            self,
            *,
            dataflow_graph_id: builtins.str | None = ...,
        ) -> None: ...
        def HasField(
            self,
            field_name: typing_extensions.Literal[
                "_dataflow_graph_id",
                b"_dataflow_graph_id",
                "dataflow_graph_id",
                b"dataflow_graph_id",
            ],
        ) -> builtins.bool: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal[
                "_dataflow_graph_id",
                b"_dataflow_graph_id",
                "dataflow_graph_id",
                b"dataflow_graph_id",
            ],
        ) -> None: ...
        def WhichOneof(
            self,
            oneof_group: typing_extensions.Literal["_dataflow_graph_id", b"_dataflow_graph_id"],
        ) -> typing_extensions.Literal["dataflow_graph_id"] | None: ...

    class DefineOutput(google.protobuf.message.Message):
        """Request to define an output: a table, a materialized view, a temporary view or a sink."""

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        class TableDetails(google.protobuf.message.Message):
            """Metadata that's only applicable to tables and materialized views."""

            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            class TablePropertiesEntry(google.protobuf.message.Message):
                DESCRIPTOR: google.protobuf.descriptor.Descriptor

                KEY_FIELD_NUMBER: builtins.int
                VALUE_FIELD_NUMBER: builtins.int
                key: builtins.str
                value: builtins.str
                def __init__(
                    self,
                    *,
                    key: builtins.str = ...,
                    value: builtins.str = ...,
                ) -> None: ...
                def ClearField(
                    self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]
                ) -> None: ...

            TABLE_PROPERTIES_FIELD_NUMBER: builtins.int
            PARTITION_COLS_FIELD_NUMBER: builtins.int
            FORMAT_FIELD_NUMBER: builtins.int
            SCHEMA_DATA_TYPE_FIELD_NUMBER: builtins.int
            SCHEMA_STRING_FIELD_NUMBER: builtins.int
            CLUSTERING_COLUMNS_FIELD_NUMBER: builtins.int
            @property
            def table_properties(
                self,
            ) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
                """Optional table properties."""
            @property
            def partition_cols(
                self,
            ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
                """Optional partition columns for the table."""
            format: builtins.str
            """The output table format for the table."""
            @property
            def schema_data_type(self) -> pyspark.sql.connect.proto.types_pb2.DataType: ...
            schema_string: builtins.str
            @property
            def clustering_columns(
                self,
            ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
                """Optional cluster columns for the table."""
            def __init__(
                self,
                *,
                table_properties: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
                partition_cols: collections.abc.Iterable[builtins.str] | None = ...,
                format: builtins.str | None = ...,
                schema_data_type: pyspark.sql.connect.proto.types_pb2.DataType | None = ...,
                schema_string: builtins.str = ...,
                clustering_columns: collections.abc.Iterable[builtins.str] | None = ...,
            ) -> None: ...
            def HasField(
                self,
                field_name: typing_extensions.Literal[
                    "_format",
                    b"_format",
                    "format",
                    b"format",
                    "schema",
                    b"schema",
                    "schema_data_type",
                    b"schema_data_type",
                    "schema_string",
                    b"schema_string",
                ],
            ) -> builtins.bool: ...
            def ClearField(
                self,
                field_name: typing_extensions.Literal[
                    "_format",
                    b"_format",
                    "clustering_columns",
                    b"clustering_columns",
                    "format",
                    b"format",
                    "partition_cols",
                    b"partition_cols",
                    "schema",
                    b"schema",
                    "schema_data_type",
                    b"schema_data_type",
                    "schema_string",
                    b"schema_string",
                    "table_properties",
                    b"table_properties",
                ],
            ) -> None: ...
            @typing.overload
            def WhichOneof(
                self, oneof_group: typing_extensions.Literal["_format", b"_format"]
            ) -> typing_extensions.Literal["format"] | None: ...
            @typing.overload
            def WhichOneof(
                self, oneof_group: typing_extensions.Literal["schema", b"schema"]
            ) -> typing_extensions.Literal["schema_data_type", "schema_string"] | None: ...

        class SinkDetails(google.protobuf.message.Message):
            """Metadata that's only applicable to sinks."""

            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            class OptionsEntry(google.protobuf.message.Message):
                DESCRIPTOR: google.protobuf.descriptor.Descriptor

                KEY_FIELD_NUMBER: builtins.int
                VALUE_FIELD_NUMBER: builtins.int
                key: builtins.str
                value: builtins.str
                def __init__(
                    self,
                    *,
                    key: builtins.str = ...,
                    value: builtins.str = ...,
                ) -> None: ...
                def ClearField(
                    self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]
                ) -> None: ...

            OPTIONS_FIELD_NUMBER: builtins.int
            FORMAT_FIELD_NUMBER: builtins.int
            @property
            def options(
                self,
            ) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
                """Streaming write options"""
            format: builtins.str
            """Streaming write format"""
            def __init__(
                self,
                *,
                options: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
                format: builtins.str | None = ...,
            ) -> None: ...
            def HasField(
                self,
                field_name: typing_extensions.Literal["_format", b"_format", "format", b"format"],
            ) -> builtins.bool: ...
            def ClearField(
                self,
                field_name: typing_extensions.Literal[
                    "_format", b"_format", "format", b"format", "options", b"options"
                ],
            ) -> None: ...
            def WhichOneof(
                self, oneof_group: typing_extensions.Literal["_format", b"_format"]
            ) -> typing_extensions.Literal["format"] | None: ...

        DATAFLOW_GRAPH_ID_FIELD_NUMBER: builtins.int
        OUTPUT_NAME_FIELD_NUMBER: builtins.int
        OUTPUT_TYPE_FIELD_NUMBER: builtins.int
        COMMENT_FIELD_NUMBER: builtins.int
        SOURCE_CODE_LOCATION_FIELD_NUMBER: builtins.int
        TABLE_DETAILS_FIELD_NUMBER: builtins.int
        SINK_DETAILS_FIELD_NUMBER: builtins.int
        EXTENSION_FIELD_NUMBER: builtins.int
        dataflow_graph_id: builtins.str
        """The graph to attach this output to."""
        output_name: builtins.str
        """Name of the output. Can be partially or fully qualified."""
        output_type: global___OutputType.ValueType
        """The type of the output."""
        comment: builtins.str
        """Optional comment for the output."""
        @property
        def source_code_location(self) -> global___SourceCodeLocation:
            """The location in source code that this output was defined."""
        @property
        def table_details(self) -> global___PipelineCommand.DefineOutput.TableDetails: ...
        @property
        def sink_details(self) -> global___PipelineCommand.DefineOutput.SinkDetails: ...
        @property
        def extension(self) -> google.protobuf.any_pb2.Any: ...
        def __init__(
            self,
            *,
            dataflow_graph_id: builtins.str | None = ...,
            output_name: builtins.str | None = ...,
            output_type: global___OutputType.ValueType | None = ...,
            comment: builtins.str | None = ...,
            source_code_location: global___SourceCodeLocation | None = ...,
            table_details: global___PipelineCommand.DefineOutput.TableDetails | None = ...,
            sink_details: global___PipelineCommand.DefineOutput.SinkDetails | None = ...,
            extension: google.protobuf.any_pb2.Any | None = ...,
        ) -> None: ...
        def HasField(
            self,
            field_name: typing_extensions.Literal[
                "_comment",
                b"_comment",
                "_dataflow_graph_id",
                b"_dataflow_graph_id",
                "_output_name",
                b"_output_name",
                "_output_type",
                b"_output_type",
                "_source_code_location",
                b"_source_code_location",
                "comment",
                b"comment",
                "dataflow_graph_id",
                b"dataflow_graph_id",
                "details",
                b"details",
                "extension",
                b"extension",
                "output_name",
                b"output_name",
                "output_type",
                b"output_type",
                "sink_details",
                b"sink_details",
                "source_code_location",
                b"source_code_location",
                "table_details",
                b"table_details",
            ],
        ) -> builtins.bool: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal[
                "_comment",
                b"_comment",
                "_dataflow_graph_id",
                b"_dataflow_graph_id",
                "_output_name",
                b"_output_name",
                "_output_type",
                b"_output_type",
                "_source_code_location",
                b"_source_code_location",
                "comment",
                b"comment",
                "dataflow_graph_id",
                b"dataflow_graph_id",
                "details",
                b"details",
                "extension",
                b"extension",
                "output_name",
                b"output_name",
                "output_type",
                b"output_type",
                "sink_details",
                b"sink_details",
                "source_code_location",
                b"source_code_location",
                "table_details",
                b"table_details",
            ],
        ) -> None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_comment", b"_comment"]
        ) -> typing_extensions.Literal["comment"] | None: ...
        @typing.overload
        def WhichOneof(
            self,
            oneof_group: typing_extensions.Literal["_dataflow_graph_id", b"_dataflow_graph_id"],
        ) -> typing_extensions.Literal["dataflow_graph_id"] | None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_output_name", b"_output_name"]
        ) -> typing_extensions.Literal["output_name"] | None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_output_type", b"_output_type"]
        ) -> typing_extensions.Literal["output_type"] | None: ...
        @typing.overload
        def WhichOneof(
            self,
            oneof_group: typing_extensions.Literal[
                "_source_code_location", b"_source_code_location"
            ],
        ) -> typing_extensions.Literal["source_code_location"] | None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["details", b"details"]
        ) -> typing_extensions.Literal["table_details", "sink_details", "extension"] | None: ...

    class DefineFlow(google.protobuf.message.Message):
        """Request to define a flow targeting a dataset."""

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        class _SCDType:
            ValueType = typing.NewType("ValueType", builtins.int)
            V: typing_extensions.TypeAlias = ValueType

        class _SCDTypeEnumTypeWrapper(
            google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[
                PipelineCommand.DefineFlow._SCDType.ValueType
            ],
            builtins.type,
        ):  # noqa: F821
            DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
            SCD_TYPE_UNSPECIFIED: PipelineCommand.DefineFlow._SCDType.ValueType  # 0
            SCD_TYPE_1: PipelineCommand.DefineFlow._SCDType.ValueType  # 1

        class SCDType(_SCDType, metaclass=_SCDTypeEnumTypeWrapper):
            """SCD Type for Auto CDC target tables."""

        SCD_TYPE_UNSPECIFIED: PipelineCommand.DefineFlow.SCDType.ValueType  # 0
        SCD_TYPE_1: PipelineCommand.DefineFlow.SCDType.ValueType  # 1

        class SqlConfEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: builtins.str
            value: builtins.str
            def __init__(
                self,
                *,
                key: builtins.str = ...,
                value: builtins.str = ...,
            ) -> None: ...
            def ClearField(
                self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]
            ) -> None: ...

        class WriteRelationFlowDetails(google.protobuf.message.Message):
            """A flow that is that takes the contents of a relation and writes it to the target dataset."""

            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            RELATION_FIELD_NUMBER: builtins.int
            @property
            def relation(self) -> pyspark.sql.connect.proto.relations_pb2.Relation:
                """An unresolved relation that defines the dataset's flow. Empty if the query function
                that defines the flow cannot be analyzed at the time of flow definition.
                """
            def __init__(
                self,
                *,
                relation: pyspark.sql.connect.proto.relations_pb2.Relation | None = ...,
            ) -> None: ...
            def HasField(
                self,
                field_name: typing_extensions.Literal[
                    "_relation", b"_relation", "relation", b"relation"
                ],
            ) -> builtins.bool: ...
            def ClearField(
                self,
                field_name: typing_extensions.Literal[
                    "_relation", b"_relation", "relation", b"relation"
                ],
            ) -> None: ...
            def WhichOneof(
                self, oneof_group: typing_extensions.Literal["_relation", b"_relation"]
            ) -> typing_extensions.Literal["relation"] | None: ...

        class AutoCdcFlowDetails(google.protobuf.message.Message):
            """Details for Auto CDC flows."""

            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            SOURCE_FIELD_NUMBER: builtins.int
            KEYS_FIELD_NUMBER: builtins.int
            SEQUENCE_BY_FIELD_NUMBER: builtins.int
            APPLY_AS_DELETES_FIELD_NUMBER: builtins.int
            APPLY_AS_TRUNCATES_FIELD_NUMBER: builtins.int
            COLUMN_LIST_FIELD_NUMBER: builtins.int
            EXCEPT_COLUMN_LIST_FIELD_NUMBER: builtins.int
            STORED_AS_SCD_TYPE_FIELD_NUMBER: builtins.int
            IGNORE_NULL_UPDATES_COLUMN_LIST_FIELD_NUMBER: builtins.int
            IGNORE_NULL_UPDATES_EXCEPT_COLUMN_LIST_FIELD_NUMBER: builtins.int
            source: builtins.str
            """The name of the CDC source to stream from."""
            @property
            def keys(
                self,
            ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
                pyspark.sql.connect.proto.expressions_pb2.Expression
            ]:
                """Column(s) that uniquely identify a row in source and target data."""
            @property
            def sequence_by(self) -> pyspark.sql.connect.proto.expressions_pb2.Expression:
                """Expression to order the source data."""
            @property
            def apply_as_deletes(self) -> pyspark.sql.connect.proto.expressions_pb2.Expression:
                """Delete condition for the merged operation."""
            @property
            def apply_as_truncates(self) -> pyspark.sql.connect.proto.expressions_pb2.Expression:
                """Truncate condition for the merged operation."""
            @property
            def column_list(
                self,
            ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
                pyspark.sql.connect.proto.expressions_pb2.Expression
            ]:
                """Columns included in the output table."""
            @property
            def except_column_list(
                self,
            ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
                pyspark.sql.connect.proto.expressions_pb2.Expression
            ]:
                """Columns excluded from the output table."""
            stored_as_scd_type: global___PipelineCommand.DefineFlow.SCDType.ValueType
            """SCD Type for target table."""
            @property
            def ignore_null_updates_column_list(
                self,
            ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
                pyspark.sql.connect.proto.expressions_pb2.Expression
            ]:
                """Subset of columns to ignore null in updates."""
            @property
            def ignore_null_updates_except_column_list(
                self,
            ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
                pyspark.sql.connect.proto.expressions_pb2.Expression
            ]:
                """Subset of columns excluded from ignoring null in updates."""
            def __init__(
                self,
                *,
                source: builtins.str | None = ...,
                keys: collections.abc.Iterable[pyspark.sql.connect.proto.expressions_pb2.Expression]
                | None = ...,
                sequence_by: pyspark.sql.connect.proto.expressions_pb2.Expression | None = ...,
                apply_as_deletes: pyspark.sql.connect.proto.expressions_pb2.Expression | None = ...,
                apply_as_truncates: pyspark.sql.connect.proto.expressions_pb2.Expression
                | None = ...,
                column_list: collections.abc.Iterable[
                    pyspark.sql.connect.proto.expressions_pb2.Expression
                ]
                | None = ...,
                except_column_list: collections.abc.Iterable[
                    pyspark.sql.connect.proto.expressions_pb2.Expression
                ]
                | None = ...,
                stored_as_scd_type: global___PipelineCommand.DefineFlow.SCDType.ValueType = ...,
                ignore_null_updates_column_list: collections.abc.Iterable[
                    pyspark.sql.connect.proto.expressions_pb2.Expression
                ]
                | None = ...,
                ignore_null_updates_except_column_list: collections.abc.Iterable[
                    pyspark.sql.connect.proto.expressions_pb2.Expression
                ]
                | None = ...,
            ) -> None: ...
            def HasField(
                self,
                field_name: typing_extensions.Literal[
                    "_apply_as_deletes",
                    b"_apply_as_deletes",
                    "_apply_as_truncates",
                    b"_apply_as_truncates",
                    "_sequence_by",
                    b"_sequence_by",
                    "_source",
                    b"_source",
                    "apply_as_deletes",
                    b"apply_as_deletes",
                    "apply_as_truncates",
                    b"apply_as_truncates",
                    "sequence_by",
                    b"sequence_by",
                    "source",
                    b"source",
                ],
            ) -> builtins.bool: ...
            def ClearField(
                self,
                field_name: typing_extensions.Literal[
                    "_apply_as_deletes",
                    b"_apply_as_deletes",
                    "_apply_as_truncates",
                    b"_apply_as_truncates",
                    "_sequence_by",
                    b"_sequence_by",
                    "_source",
                    b"_source",
                    "apply_as_deletes",
                    b"apply_as_deletes",
                    "apply_as_truncates",
                    b"apply_as_truncates",
                    "column_list",
                    b"column_list",
                    "except_column_list",
                    b"except_column_list",
                    "ignore_null_updates_column_list",
                    b"ignore_null_updates_column_list",
                    "ignore_null_updates_except_column_list",
                    b"ignore_null_updates_except_column_list",
                    "keys",
                    b"keys",
                    "sequence_by",
                    b"sequence_by",
                    "source",
                    b"source",
                    "stored_as_scd_type",
                    b"stored_as_scd_type",
                ],
            ) -> None: ...
            @typing.overload
            def WhichOneof(
                self,
                oneof_group: typing_extensions.Literal["_apply_as_deletes", b"_apply_as_deletes"],
            ) -> typing_extensions.Literal["apply_as_deletes"] | None: ...
            @typing.overload
            def WhichOneof(
                self,
                oneof_group: typing_extensions.Literal[
                    "_apply_as_truncates", b"_apply_as_truncates"
                ],
            ) -> typing_extensions.Literal["apply_as_truncates"] | None: ...
            @typing.overload
            def WhichOneof(
                self, oneof_group: typing_extensions.Literal["_sequence_by", b"_sequence_by"]
            ) -> typing_extensions.Literal["sequence_by"] | None: ...
            @typing.overload
            def WhichOneof(
                self, oneof_group: typing_extensions.Literal["_source", b"_source"]
            ) -> typing_extensions.Literal["source"] | None: ...

        class Response(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            FLOW_NAME_FIELD_NUMBER: builtins.int
            flow_name: builtins.str
            """Fully qualified flow name that uniquely identify a flow in the Dataflow graph."""
            def __init__(
                self,
                *,
                flow_name: builtins.str | None = ...,
            ) -> None: ...
            def HasField(
                self,
                field_name: typing_extensions.Literal[
                    "_flow_name", b"_flow_name", "flow_name", b"flow_name"
                ],
            ) -> builtins.bool: ...
            def ClearField(
                self,
                field_name: typing_extensions.Literal[
                    "_flow_name", b"_flow_name", "flow_name", b"flow_name"
                ],
            ) -> None: ...
            def WhichOneof(
                self, oneof_group: typing_extensions.Literal["_flow_name", b"_flow_name"]
            ) -> typing_extensions.Literal["flow_name"] | None: ...

        DATAFLOW_GRAPH_ID_FIELD_NUMBER: builtins.int
        FLOW_NAME_FIELD_NUMBER: builtins.int
        TARGET_DATASET_NAME_FIELD_NUMBER: builtins.int
        SQL_CONF_FIELD_NUMBER: builtins.int
        CLIENT_ID_FIELD_NUMBER: builtins.int
        SOURCE_CODE_LOCATION_FIELD_NUMBER: builtins.int
        RELATION_FLOW_DETAILS_FIELD_NUMBER: builtins.int
        AUTO_CDC_FLOW_DETAILS_FIELD_NUMBER: builtins.int
        EXTENSION_FIELD_NUMBER: builtins.int
        ONCE_FIELD_NUMBER: builtins.int
        dataflow_graph_id: builtins.str
        """The graph to attach this flow to."""
        flow_name: builtins.str
        """Name of the flow. For standalone flows, this must be a single-part name."""
        target_dataset_name: builtins.str
        """Name of the dataset this flow writes to. Can be partially or fully qualified."""
        @property
        def sql_conf(
            self,
        ) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
            """SQL configurations set when running this flow."""
        client_id: builtins.str
        """Identifier for the client making the request. The server uses this to determine what flow
        evaluation request stream to dispatch evaluation requests to for this flow.
        """
        @property
        def source_code_location(self) -> global___SourceCodeLocation:
            """The location in source code that this flow was defined."""
        @property
        def relation_flow_details(
            self,
        ) -> global___PipelineCommand.DefineFlow.WriteRelationFlowDetails: ...
        @property
        def auto_cdc_flow_details(
            self,
        ) -> global___PipelineCommand.DefineFlow.AutoCdcFlowDetails: ...
        @property
        def extension(self) -> google.protobuf.any_pb2.Any: ...
        once: builtins.bool
        """If true, define the flow as a one-time flow, such as for backfill.
        Set to true changes the flow in two ways:
          - The flow is run one time by default. If the pipeline is ran with a full refresh,
            the flow will run again.
          - The flow function must be a batch DataFrame, not a streaming DataFrame.
        """
        def __init__(
            self,
            *,
            dataflow_graph_id: builtins.str | None = ...,
            flow_name: builtins.str | None = ...,
            target_dataset_name: builtins.str | None = ...,
            sql_conf: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
            client_id: builtins.str | None = ...,
            source_code_location: global___SourceCodeLocation | None = ...,
            relation_flow_details: global___PipelineCommand.DefineFlow.WriteRelationFlowDetails
            | None = ...,
            auto_cdc_flow_details: global___PipelineCommand.DefineFlow.AutoCdcFlowDetails
            | None = ...,
            extension: google.protobuf.any_pb2.Any | None = ...,
            once: builtins.bool | None = ...,
        ) -> None: ...
        def HasField(
            self,
            field_name: typing_extensions.Literal[
                "_client_id",
                b"_client_id",
                "_dataflow_graph_id",
                b"_dataflow_graph_id",
                "_flow_name",
                b"_flow_name",
                "_once",
                b"_once",
                "_source_code_location",
                b"_source_code_location",
                "_target_dataset_name",
                b"_target_dataset_name",
                "auto_cdc_flow_details",
                b"auto_cdc_flow_details",
                "client_id",
                b"client_id",
                "dataflow_graph_id",
                b"dataflow_graph_id",
                "details",
                b"details",
                "extension",
                b"extension",
                "flow_name",
                b"flow_name",
                "once",
                b"once",
                "relation_flow_details",
                b"relation_flow_details",
                "source_code_location",
                b"source_code_location",
                "target_dataset_name",
                b"target_dataset_name",
            ],
        ) -> builtins.bool: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal[
                "_client_id",
                b"_client_id",
                "_dataflow_graph_id",
                b"_dataflow_graph_id",
                "_flow_name",
                b"_flow_name",
                "_once",
                b"_once",
                "_source_code_location",
                b"_source_code_location",
                "_target_dataset_name",
                b"_target_dataset_name",
                "auto_cdc_flow_details",
                b"auto_cdc_flow_details",
                "client_id",
                b"client_id",
                "dataflow_graph_id",
                b"dataflow_graph_id",
                "details",
                b"details",
                "extension",
                b"extension",
                "flow_name",
                b"flow_name",
                "once",
                b"once",
                "relation_flow_details",
                b"relation_flow_details",
                "source_code_location",
                b"source_code_location",
                "sql_conf",
                b"sql_conf",
                "target_dataset_name",
                b"target_dataset_name",
            ],
        ) -> None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_client_id", b"_client_id"]
        ) -> typing_extensions.Literal["client_id"] | None: ...
        @typing.overload
        def WhichOneof(
            self,
            oneof_group: typing_extensions.Literal["_dataflow_graph_id", b"_dataflow_graph_id"],
        ) -> typing_extensions.Literal["dataflow_graph_id"] | None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_flow_name", b"_flow_name"]
        ) -> typing_extensions.Literal["flow_name"] | None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_once", b"_once"]
        ) -> typing_extensions.Literal["once"] | None: ...
        @typing.overload
        def WhichOneof(
            self,
            oneof_group: typing_extensions.Literal[
                "_source_code_location", b"_source_code_location"
            ],
        ) -> typing_extensions.Literal["source_code_location"] | None: ...
        @typing.overload
        def WhichOneof(
            self,
            oneof_group: typing_extensions.Literal["_target_dataset_name", b"_target_dataset_name"],
        ) -> typing_extensions.Literal["target_dataset_name"] | None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["details", b"details"]
        ) -> (
            typing_extensions.Literal["relation_flow_details", "auto_cdc_flow_details", "extension"]
            | None
        ): ...

    class ExecuteOutputFlows(google.protobuf.message.Message):
        """Request to execute all flows for a single output (dataset or sink) remotely."""

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        DEFINE_OUTPUT_FIELD_NUMBER: builtins.int
        DEFINE_FLOWS_FIELD_NUMBER: builtins.int
        FULL_REFRESH_FIELD_NUMBER: builtins.int
        STORAGE_FIELD_NUMBER: builtins.int
        EXTENSION_FIELD_NUMBER: builtins.int
        @property
        def define_output(self) -> global___PipelineCommand.DefineOutput:
            """The output (table or materialized view or sink) definition."""
        @property
        def define_flows(
            self,
        ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
            global___PipelineCommand.DefineFlow
        ]:
            """The flows to execute for this table."""
        full_refresh: builtins.bool
        """Whether to perform a full refresh instead of an incremental update."""
        storage: builtins.str
        """Storage location for pipeline checkpoints and metadata."""
        @property
        def extension(
            self,
        ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
            google.protobuf.any_pb2.Any
        ]:
            """Reserved field for protocol extensions."""
        def __init__(
            self,
            *,
            define_output: global___PipelineCommand.DefineOutput | None = ...,
            define_flows: collections.abc.Iterable[global___PipelineCommand.DefineFlow]
            | None = ...,
            full_refresh: builtins.bool | None = ...,
            storage: builtins.str | None = ...,
            extension: collections.abc.Iterable[google.protobuf.any_pb2.Any] | None = ...,
        ) -> None: ...
        def HasField(
            self,
            field_name: typing_extensions.Literal[
                "_define_output",
                b"_define_output",
                "_full_refresh",
                b"_full_refresh",
                "_storage",
                b"_storage",
                "define_output",
                b"define_output",
                "full_refresh",
                b"full_refresh",
                "storage",
                b"storage",
            ],
        ) -> builtins.bool: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal[
                "_define_output",
                b"_define_output",
                "_full_refresh",
                b"_full_refresh",
                "_storage",
                b"_storage",
                "define_flows",
                b"define_flows",
                "define_output",
                b"define_output",
                "extension",
                b"extension",
                "full_refresh",
                b"full_refresh",
                "storage",
                b"storage",
            ],
        ) -> None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_define_output", b"_define_output"]
        ) -> typing_extensions.Literal["define_output"] | None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_full_refresh", b"_full_refresh"]
        ) -> typing_extensions.Literal["full_refresh"] | None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_storage", b"_storage"]
        ) -> typing_extensions.Literal["storage"] | None: ...

    class StartRun(google.protobuf.message.Message):
        """Resolves all datasets and flows and start a pipeline update. Should be called after all
        graph elements are registered.
        """

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        DATAFLOW_GRAPH_ID_FIELD_NUMBER: builtins.int
        FULL_REFRESH_SELECTION_FIELD_NUMBER: builtins.int
        FULL_REFRESH_ALL_FIELD_NUMBER: builtins.int
        REFRESH_SELECTION_FIELD_NUMBER: builtins.int
        DRY_FIELD_NUMBER: builtins.int
        STORAGE_FIELD_NUMBER: builtins.int
        dataflow_graph_id: builtins.str
        """The graph to start."""
        @property
        def full_refresh_selection(
            self,
        ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
            """List of dataset to reset and recompute."""
        full_refresh_all: builtins.bool
        """Perform a full graph reset and recompute."""
        @property
        def refresh_selection(
            self,
        ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
            """List of dataset to update."""
        dry: builtins.bool
        """If true, the run will not actually execute any flows, but will only validate the graph and
        check for any errors. This is useful for testing and validation purposes.
        """
        storage: builtins.str
        """storage location for pipeline checkpoints and metadata."""
        def __init__(
            self,
            *,
            dataflow_graph_id: builtins.str | None = ...,
            full_refresh_selection: collections.abc.Iterable[builtins.str] | None = ...,
            full_refresh_all: builtins.bool | None = ...,
            refresh_selection: collections.abc.Iterable[builtins.str] | None = ...,
            dry: builtins.bool | None = ...,
            storage: builtins.str | None = ...,
        ) -> None: ...
        def HasField(
            self,
            field_name: typing_extensions.Literal[
                "_dataflow_graph_id",
                b"_dataflow_graph_id",
                "_dry",
                b"_dry",
                "_full_refresh_all",
                b"_full_refresh_all",
                "_storage",
                b"_storage",
                "dataflow_graph_id",
                b"dataflow_graph_id",
                "dry",
                b"dry",
                "full_refresh_all",
                b"full_refresh_all",
                "storage",
                b"storage",
            ],
        ) -> builtins.bool: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal[
                "_dataflow_graph_id",
                b"_dataflow_graph_id",
                "_dry",
                b"_dry",
                "_full_refresh_all",
                b"_full_refresh_all",
                "_storage",
                b"_storage",
                "dataflow_graph_id",
                b"dataflow_graph_id",
                "dry",
                b"dry",
                "full_refresh_all",
                b"full_refresh_all",
                "full_refresh_selection",
                b"full_refresh_selection",
                "refresh_selection",
                b"refresh_selection",
                "storage",
                b"storage",
            ],
        ) -> None: ...
        @typing.overload
        def WhichOneof(
            self,
            oneof_group: typing_extensions.Literal["_dataflow_graph_id", b"_dataflow_graph_id"],
        ) -> typing_extensions.Literal["dataflow_graph_id"] | None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_dry", b"_dry"]
        ) -> typing_extensions.Literal["dry"] | None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_full_refresh_all", b"_full_refresh_all"]
        ) -> typing_extensions.Literal["full_refresh_all"] | None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_storage", b"_storage"]
        ) -> typing_extensions.Literal["storage"] | None: ...

    class DefineSqlGraphElements(google.protobuf.message.Message):
        """Parses the SQL file and registers all datasets and flows."""

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        DATAFLOW_GRAPH_ID_FIELD_NUMBER: builtins.int
        SQL_FILE_PATH_FIELD_NUMBER: builtins.int
        SQL_TEXT_FIELD_NUMBER: builtins.int
        dataflow_graph_id: builtins.str
        """The graph to attach this dataset to."""
        sql_file_path: builtins.str
        """The full path to the SQL file. Can be relative or absolute."""
        sql_text: builtins.str
        """The contents of the SQL file."""
        def __init__(
            self,
            *,
            dataflow_graph_id: builtins.str | None = ...,
            sql_file_path: builtins.str | None = ...,
            sql_text: builtins.str | None = ...,
        ) -> None: ...
        def HasField(
            self,
            field_name: typing_extensions.Literal[
                "_dataflow_graph_id",
                b"_dataflow_graph_id",
                "_sql_file_path",
                b"_sql_file_path",
                "_sql_text",
                b"_sql_text",
                "dataflow_graph_id",
                b"dataflow_graph_id",
                "sql_file_path",
                b"sql_file_path",
                "sql_text",
                b"sql_text",
            ],
        ) -> builtins.bool: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal[
                "_dataflow_graph_id",
                b"_dataflow_graph_id",
                "_sql_file_path",
                b"_sql_file_path",
                "_sql_text",
                b"_sql_text",
                "dataflow_graph_id",
                b"dataflow_graph_id",
                "sql_file_path",
                b"sql_file_path",
                "sql_text",
                b"sql_text",
            ],
        ) -> None: ...
        @typing.overload
        def WhichOneof(
            self,
            oneof_group: typing_extensions.Literal["_dataflow_graph_id", b"_dataflow_graph_id"],
        ) -> typing_extensions.Literal["dataflow_graph_id"] | None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_sql_file_path", b"_sql_file_path"]
        ) -> typing_extensions.Literal["sql_file_path"] | None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_sql_text", b"_sql_text"]
        ) -> typing_extensions.Literal["sql_text"] | None: ...

    class GetQueryFunctionExecutionSignalStream(google.protobuf.message.Message):
        """Request to get the stream of query function execution signals for a graph. Responses should
        be a stream of PipelineQueryFunctionExecutionSignal messages.
        """

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        DATAFLOW_GRAPH_ID_FIELD_NUMBER: builtins.int
        CLIENT_ID_FIELD_NUMBER: builtins.int
        dataflow_graph_id: builtins.str
        """The graph to get the query function execution signal stream for."""
        client_id: builtins.str
        """Identifier for the client that is requesting the stream."""
        def __init__(
            self,
            *,
            dataflow_graph_id: builtins.str | None = ...,
            client_id: builtins.str | None = ...,
        ) -> None: ...
        def HasField(
            self,
            field_name: typing_extensions.Literal[
                "_client_id",
                b"_client_id",
                "_dataflow_graph_id",
                b"_dataflow_graph_id",
                "client_id",
                b"client_id",
                "dataflow_graph_id",
                b"dataflow_graph_id",
            ],
        ) -> builtins.bool: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal[
                "_client_id",
                b"_client_id",
                "_dataflow_graph_id",
                b"_dataflow_graph_id",
                "client_id",
                b"client_id",
                "dataflow_graph_id",
                b"dataflow_graph_id",
            ],
        ) -> None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_client_id", b"_client_id"]
        ) -> typing_extensions.Literal["client_id"] | None: ...
        @typing.overload
        def WhichOneof(
            self,
            oneof_group: typing_extensions.Literal["_dataflow_graph_id", b"_dataflow_graph_id"],
        ) -> typing_extensions.Literal["dataflow_graph_id"] | None: ...

    class DefineFlowQueryFunctionResult(google.protobuf.message.Message):
        """Request from the client to update the flow function evaluation result
        for a previously un-analyzed flow.
        """

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        FLOW_NAME_FIELD_NUMBER: builtins.int
        FLOW_IDENTIFIER_FIELD_NUMBER: builtins.int
        DATAFLOW_GRAPH_ID_FIELD_NUMBER: builtins.int
        RELATION_FIELD_NUMBER: builtins.int
        flow_name: builtins.str
        """(Deprecated) The fully qualified name of the flow being updated.

        This field is deprecated since Spark 4.2+. Use flow_identifier field instead.
        """
        @property
        def flow_identifier(self) -> pyspark.sql.connect.proto.common_pb2.ResolvedIdentifier:
            """The fully qualified identifier of the flow being updated."""
        dataflow_graph_id: builtins.str
        """The ID of the graph this flow belongs to."""
        @property
        def relation(self) -> pyspark.sql.connect.proto.relations_pb2.Relation:
            """An unresolved relation that defines the dataset's flow."""
        def __init__(
            self,
            *,
            flow_name: builtins.str | None = ...,
            flow_identifier: pyspark.sql.connect.proto.common_pb2.ResolvedIdentifier | None = ...,
            dataflow_graph_id: builtins.str | None = ...,
            relation: pyspark.sql.connect.proto.relations_pb2.Relation | None = ...,
        ) -> None: ...
        def HasField(
            self,
            field_name: typing_extensions.Literal[
                "_dataflow_graph_id",
                b"_dataflow_graph_id",
                "_flow_identifier",
                b"_flow_identifier",
                "_flow_name",
                b"_flow_name",
                "_relation",
                b"_relation",
                "dataflow_graph_id",
                b"dataflow_graph_id",
                "flow_identifier",
                b"flow_identifier",
                "flow_name",
                b"flow_name",
                "relation",
                b"relation",
            ],
        ) -> builtins.bool: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal[
                "_dataflow_graph_id",
                b"_dataflow_graph_id",
                "_flow_identifier",
                b"_flow_identifier",
                "_flow_name",
                b"_flow_name",
                "_relation",
                b"_relation",
                "dataflow_graph_id",
                b"dataflow_graph_id",
                "flow_identifier",
                b"flow_identifier",
                "flow_name",
                b"flow_name",
                "relation",
                b"relation",
            ],
        ) -> None: ...
        @typing.overload
        def WhichOneof(
            self,
            oneof_group: typing_extensions.Literal["_dataflow_graph_id", b"_dataflow_graph_id"],
        ) -> typing_extensions.Literal["dataflow_graph_id"] | None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_flow_identifier", b"_flow_identifier"]
        ) -> typing_extensions.Literal["flow_identifier"] | None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_flow_name", b"_flow_name"]
        ) -> typing_extensions.Literal["flow_name"] | None: ...
        @typing.overload
        def WhichOneof(
            self, oneof_group: typing_extensions.Literal["_relation", b"_relation"]
        ) -> typing_extensions.Literal["relation"] | None: ...

    CREATE_DATAFLOW_GRAPH_FIELD_NUMBER: builtins.int
    DEFINE_OUTPUT_FIELD_NUMBER: builtins.int
    DEFINE_FLOW_FIELD_NUMBER: builtins.int
    DROP_DATAFLOW_GRAPH_FIELD_NUMBER: builtins.int
    START_RUN_FIELD_NUMBER: builtins.int
    DEFINE_SQL_GRAPH_ELEMENTS_FIELD_NUMBER: builtins.int
    GET_QUERY_FUNCTION_EXECUTION_SIGNAL_STREAM_FIELD_NUMBER: builtins.int
    DEFINE_FLOW_QUERY_FUNCTION_RESULT_FIELD_NUMBER: builtins.int
    EXECUTE_OUTPUT_FLOWS_FIELD_NUMBER: builtins.int
    EXTENSION_FIELD_NUMBER: builtins.int
    @property
    def create_dataflow_graph(self) -> global___PipelineCommand.CreateDataflowGraph: ...
    @property
    def define_output(self) -> global___PipelineCommand.DefineOutput: ...
    @property
    def define_flow(self) -> global___PipelineCommand.DefineFlow: ...
    @property
    def drop_dataflow_graph(self) -> global___PipelineCommand.DropDataflowGraph: ...
    @property
    def start_run(self) -> global___PipelineCommand.StartRun: ...
    @property
    def define_sql_graph_elements(self) -> global___PipelineCommand.DefineSqlGraphElements: ...
    @property
    def get_query_function_execution_signal_stream(
        self,
    ) -> global___PipelineCommand.GetQueryFunctionExecutionSignalStream: ...
    @property
    def define_flow_query_function_result(
        self,
    ) -> global___PipelineCommand.DefineFlowQueryFunctionResult: ...
    @property
    def execute_output_flows(self) -> global___PipelineCommand.ExecuteOutputFlows: ...
    @property
    def extension(self) -> google.protobuf.any_pb2.Any:
        """Reserved field for protocol extensions.
        Used to support forward-compatibility by carrying additional command types
        that are not yet defined in this version of the proto. During planning, the
        engine will resolve and dispatch the concrete command contained in this field.
        """
    def __init__(
        self,
        *,
        create_dataflow_graph: global___PipelineCommand.CreateDataflowGraph | None = ...,
        define_output: global___PipelineCommand.DefineOutput | None = ...,
        define_flow: global___PipelineCommand.DefineFlow | None = ...,
        drop_dataflow_graph: global___PipelineCommand.DropDataflowGraph | None = ...,
        start_run: global___PipelineCommand.StartRun | None = ...,
        define_sql_graph_elements: global___PipelineCommand.DefineSqlGraphElements | None = ...,
        get_query_function_execution_signal_stream: global___PipelineCommand.GetQueryFunctionExecutionSignalStream
        | None = ...,
        define_flow_query_function_result: global___PipelineCommand.DefineFlowQueryFunctionResult
        | None = ...,
        execute_output_flows: global___PipelineCommand.ExecuteOutputFlows | None = ...,
        extension: google.protobuf.any_pb2.Any | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "command_type",
            b"command_type",
            "create_dataflow_graph",
            b"create_dataflow_graph",
            "define_flow",
            b"define_flow",
            "define_flow_query_function_result",
            b"define_flow_query_function_result",
            "define_output",
            b"define_output",
            "define_sql_graph_elements",
            b"define_sql_graph_elements",
            "drop_dataflow_graph",
            b"drop_dataflow_graph",
            "execute_output_flows",
            b"execute_output_flows",
            "extension",
            b"extension",
            "get_query_function_execution_signal_stream",
            b"get_query_function_execution_signal_stream",
            "start_run",
            b"start_run",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "command_type",
            b"command_type",
            "create_dataflow_graph",
            b"create_dataflow_graph",
            "define_flow",
            b"define_flow",
            "define_flow_query_function_result",
            b"define_flow_query_function_result",
            "define_output",
            b"define_output",
            "define_sql_graph_elements",
            b"define_sql_graph_elements",
            "drop_dataflow_graph",
            b"drop_dataflow_graph",
            "execute_output_flows",
            b"execute_output_flows",
            "extension",
            b"extension",
            "get_query_function_execution_signal_stream",
            b"get_query_function_execution_signal_stream",
            "start_run",
            b"start_run",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["command_type", b"command_type"]
    ) -> (
        typing_extensions.Literal[
            "create_dataflow_graph",
            "define_output",
            "define_flow",
            "drop_dataflow_graph",
            "start_run",
            "define_sql_graph_elements",
            "get_query_function_execution_signal_stream",
            "define_flow_query_function_result",
            "execute_output_flows",
            "extension",
        ]
        | None
    ): ...

global___PipelineCommand = PipelineCommand

class PipelineCommandResult(google.protobuf.message.Message):
    """Dispatch object for pipelines command results."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class CreateDataflowGraphResult(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        DATAFLOW_GRAPH_ID_FIELD_NUMBER: builtins.int
        dataflow_graph_id: builtins.str
        """The ID of the created graph."""
        def __init__(
            self,
            *,
            dataflow_graph_id: builtins.str | None = ...,
        ) -> None: ...
        def HasField(
            self,
            field_name: typing_extensions.Literal[
                "_dataflow_graph_id",
                b"_dataflow_graph_id",
                "dataflow_graph_id",
                b"dataflow_graph_id",
            ],
        ) -> builtins.bool: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal[
                "_dataflow_graph_id",
                b"_dataflow_graph_id",
                "dataflow_graph_id",
                b"dataflow_graph_id",
            ],
        ) -> None: ...
        def WhichOneof(
            self,
            oneof_group: typing_extensions.Literal["_dataflow_graph_id", b"_dataflow_graph_id"],
        ) -> typing_extensions.Literal["dataflow_graph_id"] | None: ...

    class DefineOutputResult(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        RESOLVED_IDENTIFIER_FIELD_NUMBER: builtins.int
        @property
        def resolved_identifier(self) -> pyspark.sql.connect.proto.common_pb2.ResolvedIdentifier:
            """Resolved identifier of the output"""
        def __init__(
            self,
            *,
            resolved_identifier: pyspark.sql.connect.proto.common_pb2.ResolvedIdentifier
            | None = ...,
        ) -> None: ...
        def HasField(
            self,
            field_name: typing_extensions.Literal[
                "_resolved_identifier",
                b"_resolved_identifier",
                "resolved_identifier",
                b"resolved_identifier",
            ],
        ) -> builtins.bool: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal[
                "_resolved_identifier",
                b"_resolved_identifier",
                "resolved_identifier",
                b"resolved_identifier",
            ],
        ) -> None: ...
        def WhichOneof(
            self,
            oneof_group: typing_extensions.Literal["_resolved_identifier", b"_resolved_identifier"],
        ) -> typing_extensions.Literal["resolved_identifier"] | None: ...

    class DefineFlowResult(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        RESOLVED_IDENTIFIER_FIELD_NUMBER: builtins.int
        @property
        def resolved_identifier(self) -> pyspark.sql.connect.proto.common_pb2.ResolvedIdentifier:
            """Resolved identifier of the flow"""
        def __init__(
            self,
            *,
            resolved_identifier: pyspark.sql.connect.proto.common_pb2.ResolvedIdentifier
            | None = ...,
        ) -> None: ...
        def HasField(
            self,
            field_name: typing_extensions.Literal[
                "_resolved_identifier",
                b"_resolved_identifier",
                "resolved_identifier",
                b"resolved_identifier",
            ],
        ) -> builtins.bool: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal[
                "_resolved_identifier",
                b"_resolved_identifier",
                "resolved_identifier",
                b"resolved_identifier",
            ],
        ) -> None: ...
        def WhichOneof(
            self,
            oneof_group: typing_extensions.Literal["_resolved_identifier", b"_resolved_identifier"],
        ) -> typing_extensions.Literal["resolved_identifier"] | None: ...

    CREATE_DATAFLOW_GRAPH_RESULT_FIELD_NUMBER: builtins.int
    DEFINE_OUTPUT_RESULT_FIELD_NUMBER: builtins.int
    DEFINE_FLOW_RESULT_FIELD_NUMBER: builtins.int
    @property
    def create_dataflow_graph_result(
        self,
    ) -> global___PipelineCommandResult.CreateDataflowGraphResult: ...
    @property
    def define_output_result(self) -> global___PipelineCommandResult.DefineOutputResult: ...
    @property
    def define_flow_result(self) -> global___PipelineCommandResult.DefineFlowResult: ...
    def __init__(
        self,
        *,
        create_dataflow_graph_result: global___PipelineCommandResult.CreateDataflowGraphResult
        | None = ...,
        define_output_result: global___PipelineCommandResult.DefineOutputResult | None = ...,
        define_flow_result: global___PipelineCommandResult.DefineFlowResult | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "create_dataflow_graph_result",
            b"create_dataflow_graph_result",
            "define_flow_result",
            b"define_flow_result",
            "define_output_result",
            b"define_output_result",
            "result_type",
            b"result_type",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "create_dataflow_graph_result",
            b"create_dataflow_graph_result",
            "define_flow_result",
            b"define_flow_result",
            "define_output_result",
            b"define_output_result",
            "result_type",
            b"result_type",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["result_type", b"result_type"]
    ) -> (
        typing_extensions.Literal[
            "create_dataflow_graph_result", "define_output_result", "define_flow_result"
        ]
        | None
    ): ...

global___PipelineCommandResult = PipelineCommandResult

class PipelineEventResult(google.protobuf.message.Message):
    """A response containing an event emitted during the run of a pipeline."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    EVENT_FIELD_NUMBER: builtins.int
    @property
    def event(self) -> global___PipelineEvent: ...
    def __init__(
        self,
        *,
        event: global___PipelineEvent | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing_extensions.Literal["event", b"event"]
    ) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["event", b"event"]) -> None: ...

global___PipelineEventResult = PipelineEventResult

class PipelineEvent(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TIMESTAMP_FIELD_NUMBER: builtins.int
    MESSAGE_FIELD_NUMBER: builtins.int
    @property
    def timestamp(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """The timestamp corresponding to when the event occurred."""
    message: builtins.str
    """The message that should be displayed to users."""
    def __init__(
        self,
        *,
        timestamp: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        message: builtins.str | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "_message", b"_message", "message", b"message", "timestamp", b"timestamp"
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "_message", b"_message", "message", b"message", "timestamp", b"timestamp"
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["_message", b"_message"]
    ) -> typing_extensions.Literal["message"] | None: ...

global___PipelineEvent = PipelineEvent

class SourceCodeLocation(google.protobuf.message.Message):
    """Source code location information associated with a particular dataset or flow."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FILE_NAME_FIELD_NUMBER: builtins.int
    LINE_NUMBER_FIELD_NUMBER: builtins.int
    DEFINITION_PATH_FIELD_NUMBER: builtins.int
    EXTENSION_FIELD_NUMBER: builtins.int
    file_name: builtins.str
    """The file that this pipeline source code was defined in."""
    line_number: builtins.int
    """The specific line number that this pipeline source code is located at, if applicable."""
    definition_path: builtins.str
    """The path of the top-level pipeline file determined at runtime during pipeline initialization."""
    @property
    def extension(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        google.protobuf.any_pb2.Any
    ]:
        """Reserved field for protocol extensions.
        Used to support forward-compatibility by carrying additional fields
        that are not yet defined in this version of the proto. During planning, the
        engine will resolve and dispatch the concrete command contained in this field.
        """
    def __init__(
        self,
        *,
        file_name: builtins.str | None = ...,
        line_number: builtins.int | None = ...,
        definition_path: builtins.str | None = ...,
        extension: collections.abc.Iterable[google.protobuf.any_pb2.Any] | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "_definition_path",
            b"_definition_path",
            "_file_name",
            b"_file_name",
            "_line_number",
            b"_line_number",
            "definition_path",
            b"definition_path",
            "file_name",
            b"file_name",
            "line_number",
            b"line_number",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "_definition_path",
            b"_definition_path",
            "_file_name",
            b"_file_name",
            "_line_number",
            b"_line_number",
            "definition_path",
            b"definition_path",
            "extension",
            b"extension",
            "file_name",
            b"file_name",
            "line_number",
            b"line_number",
        ],
    ) -> None: ...
    @typing.overload
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["_definition_path", b"_definition_path"]
    ) -> typing_extensions.Literal["definition_path"] | None: ...
    @typing.overload
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["_file_name", b"_file_name"]
    ) -> typing_extensions.Literal["file_name"] | None: ...
    @typing.overload
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["_line_number", b"_line_number"]
    ) -> typing_extensions.Literal["line_number"] | None: ...

global___SourceCodeLocation = SourceCodeLocation

class PipelineQueryFunctionExecutionSignal(google.protobuf.message.Message):
    """A signal from the server to the client to execute the query function for one or more flows, and
    to register their results with the server.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FLOW_NAMES_FIELD_NUMBER: builtins.int
    FLOW_IDENTIFIERS_FIELD_NUMBER: builtins.int
    @property
    def flow_names(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """(Deprecated) The name of flows that are ready to be re-evaluated.

        This field is deprecated since Spark 4.2+. Use flow_identifiers field instead.
        """
    @property
    def flow_identifiers(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        pyspark.sql.connect.proto.common_pb2.ResolvedIdentifier
    ]:
        """The identifier of flows that are ready to be re-evaluated"""
    def __init__(
        self,
        *,
        flow_names: collections.abc.Iterable[builtins.str] | None = ...,
        flow_identifiers: collections.abc.Iterable[
            pyspark.sql.connect.proto.common_pb2.ResolvedIdentifier
        ]
        | None = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "flow_identifiers", b"flow_identifiers", "flow_names", b"flow_names"
        ],
    ) -> None: ...

global___PipelineQueryFunctionExecutionSignal = PipelineQueryFunctionExecutionSignal

class PipelineAnalysisContext(google.protobuf.message.Message):
    """Metadata providing context about the pipeline during Spark Connect query analysis."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DATAFLOW_GRAPH_ID_FIELD_NUMBER: builtins.int
    DEFINITION_PATH_FIELD_NUMBER: builtins.int
    FLOW_NAME_FIELD_NUMBER: builtins.int
    FLOW_IDENTIFIER_FIELD_NUMBER: builtins.int
    EXTENSION_FIELD_NUMBER: builtins.int
    dataflow_graph_id: builtins.str
    """Unique identifier of the dataflow graph associated with this pipeline."""
    definition_path: builtins.str
    """The path of the top-level pipeline file determined at runtime during pipeline initialization."""
    flow_name: builtins.str
    """(Deprecated) The name of the Flow involved in this analysis

    This field is deprecated since Spark 4.2+. Use flow_identifier field instead.
    """
    @property
    def flow_identifier(self) -> pyspark.sql.connect.proto.common_pb2.ResolvedIdentifier:
        """The identifier of the Flow involved in this analysis"""
    @property
    def extension(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        google.protobuf.any_pb2.Any
    ]:
        """Reserved field for protocol extensions."""
    def __init__(
        self,
        *,
        dataflow_graph_id: builtins.str | None = ...,
        definition_path: builtins.str | None = ...,
        flow_name: builtins.str | None = ...,
        flow_identifier: pyspark.sql.connect.proto.common_pb2.ResolvedIdentifier | None = ...,
        extension: collections.abc.Iterable[google.protobuf.any_pb2.Any] | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "_dataflow_graph_id",
            b"_dataflow_graph_id",
            "_definition_path",
            b"_definition_path",
            "_flow_identifier",
            b"_flow_identifier",
            "_flow_name",
            b"_flow_name",
            "dataflow_graph_id",
            b"dataflow_graph_id",
            "definition_path",
            b"definition_path",
            "flow_identifier",
            b"flow_identifier",
            "flow_name",
            b"flow_name",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "_dataflow_graph_id",
            b"_dataflow_graph_id",
            "_definition_path",
            b"_definition_path",
            "_flow_identifier",
            b"_flow_identifier",
            "_flow_name",
            b"_flow_name",
            "dataflow_graph_id",
            b"dataflow_graph_id",
            "definition_path",
            b"definition_path",
            "extension",
            b"extension",
            "flow_identifier",
            b"flow_identifier",
            "flow_name",
            b"flow_name",
        ],
    ) -> None: ...
    @typing.overload
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["_dataflow_graph_id", b"_dataflow_graph_id"]
    ) -> typing_extensions.Literal["dataflow_graph_id"] | None: ...
    @typing.overload
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["_definition_path", b"_definition_path"]
    ) -> typing_extensions.Literal["definition_path"] | None: ...
    @typing.overload
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["_flow_identifier", b"_flow_identifier"]
    ) -> typing_extensions.Literal["flow_identifier"] | None: ...
    @typing.overload
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["_flow_name", b"_flow_name"]
    ) -> typing_extensions.Literal["flow_name"] | None: ...

global___PipelineAnalysisContext = PipelineAnalysisContext



# -*- coding: utf-8 -*-
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import unittest

from pyspark.util import is_remote_only
from pyspark.testing.connectutils import should_test_connect, connect_requirement_message
from pyspark.testing.utils import have_torch, torch_requirement_message
from pyspark.testing.connectutils import ReusedConnectTestCase

if should_test_connect:
    from pyspark.ml.tests.connect.test_legacy_mode_tuning import CrossValidatorTestsMixin

    @unittest.skipIf(
        not should_test_connect or not have_torch or is_remote_only(),
        connect_requirement_message
        or torch_requirement_message
        or "Requires PySpark core library in Spark Connect server",
    )
    class CrossValidatorTestsOnConnect(CrossValidatorTestsMixin, ReusedConnectTestCase):
        @classmethod
        def conf(cls):
            config = super().conf()
            config.set("spark.sql.artifact.copyFromLocalToFs.allowDestLocal", "true")
            return config

        @classmethod
        def master(cls):
            return os.environ.get("SPARK_CONNECT_TESTING_REMOTE", "local[2]")


if __name__ == "__main__":
    from pyspark.testing import main

    main()



#!/usr/bin/env python3

#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This script generates llms.txt file for Apache Spark documentation

import sys
import argparse
from pathlib import Path


def generate_llms_txt(docs_path: Path, output_path: Path, version: str = "latest") -> None:
    """
    Generate the llms.txt file for Apache Spark documentation with hardcoded categories.
    """
    content = []
    content.append("# Apache Spark")
    content.append("")
    content.append(
        "> Apache Spark™ is a unified analytics engine for large-scale data processing. "
        "It provides high-level APIs in Java, Scala, Python, and R, and an optimized engine "
        "that supports general execution graphs. It also supports a rich set of higher-level "
        "tools including Spark SQL for SQL and structured data processing, MLlib for machine "
        "learning, GraphX for graph processing, and Structured Streaming for incremental "
        "computation and stream processing."
    )
    content.append("")

    doc_home_url = f"https://spark.apache.org/docs/{version}/"
    content.append(f"Documentation home: {doc_home_url}")
    content.append("")

    content.append("## Programming Guides")
    content.append("")
    programming_guides = [
        ("Quick Start", f"https://spark.apache.org/docs/{version}/quick-start.html"),
        ("Spark SQL", f"https://spark.apache.org/docs/{version}/sql-programming-guide.html"),
        (
            "PySpark",
            f"https://spark.apache.org/docs/{version}/api/python/getting_started/index.html",
        ),
        (
            "RDD Programming Guide",
            f"https://spark.apache.org/docs/{version}/rdd-programming-guide.html",
        ),
        (
            "Structured Streaming",
            f"https://spark.apache.org/docs/{version}/streaming-programming-guide.html",
        ),
        ("MLlib", f"https://spark.apache.org/docs/{version}/ml-guide.html"),
        ("GraphX", f"https://spark.apache.org/docs/{version}/graphx-programming-guide.html"),
        ("SparkR", f"https://spark.apache.org/docs/{version}/sparkr.html"),
        (
            "Spark SQL CLI",
            f"https://spark.apache.org/docs/{version}/"
            f"sql-distributed-sql-engine-spark-sql-cli.html",
        ),
    ]
    for title, url in programming_guides:
        content.append(f"- [{title}]({url})")
    content.append("")

    content.append("## API Docs")
    content.append("")
    # TODO: Update API docs to point to their own llms.txt files once available
    # e.g., https://spark.apache.org/docs/{version}/api/python/llms.txt
    api_docs = [
        ("Spark Python API", f"https://spark.apache.org/docs/{version}/api/python/index.html"),
        (
            "Spark Scala API",
            f"https://spark.apache.org/docs/{version}/api/scala/org/apache/spark/index.html",
        ),
        ("Spark Java API", f"https://spark.apache.org/docs/{version}/api/java/index.html"),
        ("Spark R API", f"https://spark.apache.org/docs/{version}/api/R/index.html"),
        (
            "Spark SQL Built-in Functions",
            f"https://spark.apache.org/docs/{version}/api/sql/index.html",
        ),
    ]
    for title, url in api_docs:
        content.append(f"- [{title}]({url})")
    content.append("")

    content.append("## Deployment Guides")
    content.append("")
    deployment_guides = [
        ("Cluster Overview", f"https://spark.apache.org/docs/{version}/cluster-overview.html"),
        (
            "Submitting Applications",
            f"https://spark.apache.org/docs/{version}/submitting-applications.html",
        ),
        (
            "Standalone Deploy Mode",
            f"https://spark.apache.org/docs/{version}/spark-standalone.html",
        ),
        ("YARN", f"https://spark.apache.org/docs/{version}/running-on-yarn.html"),
        ("Kubernetes", f"https://spark.apache.org/docs/{version}/running-on-kubernetes.html"),
    ]
    for title, url in deployment_guides:
        content.append(f"- [{title}]({url})")
    content.append("")

    content.append("## Other Documents")
    content.append("")
    other_docs = [
        ("Configuration", f"https://spark.apache.org/docs/{version}/configuration.html"),
        ("Monitoring", f"https://spark.apache.org/docs/{version}/monitoring.html"),
        ("Web UI", f"https://spark.apache.org/docs/{version}/web-ui.html"),
        ("Tuning Guide", f"https://spark.apache.org/docs/{version}/tuning.html"),
        ("Job Scheduling", f"https://spark.apache.org/docs/{version}/job-scheduling.html"),
        ("Security", f"https://spark.apache.org/docs/{version}/security.html"),
        (
            "Hardware Provisioning",
            f"https://spark.apache.org/docs/{version}/hardware-provisioning.html",
        ),
        (
            "Cloud Infrastructures",
            f"https://spark.apache.org/docs/{version}/cloud-integration.html",
        ),
        ("Migration Guide", f"https://spark.apache.org/docs/{version}/migration-guide.html"),
    ]
    for title, url in other_docs:
        content.append(f"- [{title}]({url})")
    content.append("")

    content.append("## External Resources")
    content.append("")
    content.append("- [Apache Spark Home](https://spark.apache.org/)")
    content.append("- [Downloads](https://spark.apache.org/downloads.html)")
    content.append("- [GitHub Repository](https://github.com/apache/spark)")
    content.append("- [Issue Tracker (JIRA)](https://issues.apache.org/jira/projects/SPARK)")
    content.append("- [Mailing Lists](https://spark.apache.org/mailing-lists.html)")
    content.append("- [Community](https://spark.apache.org/community.html)")
    content.append("- [Contributing](https://spark.apache.org/contributing.html)")
    content.append("")

    with open(output_path, "w", encoding="utf-8") as f:
        f.write("\n".join(content))

    print(f"Generated {output_path}")

    total_docs = len(programming_guides) + len(api_docs) + len(deployment_guides) + len(other_docs)
    sections_count = 5

    print(f"Total documentation pages indexed: {total_docs}")
    print(f"Sections: {sections_count}")


def main():
    parser = argparse.ArgumentParser(
        description="Generate llms.txt file for Apache Spark documentation"
    )
    parser.add_argument(
        "--docs-path", type=str, default="docs", help="Path to the docs directory (default: docs)"
    )
    parser.add_argument(
        "--output", type=str, default="llms.txt", help="Output file path (default: llms.txt)"
    )
    parser.add_argument(
        "--version",
        type=str,
        default="latest",
        help="Spark documentation version (default: latest)",
    )

    args = parser.parse_args()

    # Convert to Path objects
    script_dir = Path(__file__).parent
    project_root = script_dir.parent.parent  # Go up two levels from dev/create-release/
    docs_path = project_root / args.docs_path
    output_path = project_root / args.output

    # Check if docs directory exists
    if not docs_path.exists():
        print(f"Error: Documentation directory '{docs_path}' does not exist")
        sys.exit(1)

    # Generate the llms.txt file
    generate_llms_txt(docs_path, output_path, args.version)


if __name__ == "__main__":
    main()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Any, Union, cast

from pandas.api.types import CategoricalDtype

from pyspark.pandas._typing import Dtype, IndexOpsLike, SeriesOrIndex
from pyspark.pandas.base import column_op, IndexOpsMixin
from pyspark.pandas.data_type_ops.base import (
    DataTypeOps,
    _as_bool_type,
    _as_categorical_type,
    _as_other_type,
    _as_string_type,
    _sanitize_list_like,
)
from pyspark.pandas.typedef import pandas_on_spark_type
from pyspark.sql import functions as F, Column
from pyspark.sql.types import ArrayType, BooleanType, NumericType, StringType


class ArrayOps(DataTypeOps):
    """
    The class for binary operations of pandas-on-Spark objects with ArrayType.
    """

    @property
    def pretty_name(self) -> str:
        return "arrays"

    def add(self, left: IndexOpsLike, right: Any) -> SeriesOrIndex:
        _sanitize_list_like(right)
        if not isinstance(right, IndexOpsMixin) or (
            isinstance(right, IndexOpsMixin) and not isinstance(right.spark.data_type, ArrayType)
        ):
            raise TypeError(
                "Concatenation can not be applied to %s and the given type." % self.pretty_name
            )

        left_type = cast(ArrayType, left.spark.data_type).elementType
        right_type = right.spark.data_type.elementType

        if left_type != right_type and not (
            isinstance(left_type, NumericType) and isinstance(right_type, NumericType)
        ):
            raise TypeError(
                "Concatenation can only be applied to %s of the same type" % self.pretty_name
            )

        return column_op(F.concat)(left, right)

    def lt(self, left: IndexOpsLike, right: Any) -> SeriesOrIndex:
        from pyspark.pandas.base import column_op

        _sanitize_list_like(right)
        return column_op(Column.__lt__)(left, right)

    def le(self, left: IndexOpsLike, right: Any) -> SeriesOrIndex:
        from pyspark.pandas.base import column_op

        _sanitize_list_like(right)
        return column_op(Column.__le__)(left, right)

    def ge(self, left: IndexOpsLike, right: Any) -> SeriesOrIndex:
        from pyspark.pandas.base import column_op

        _sanitize_list_like(right)
        return column_op(Column.__ge__)(left, right)

    def gt(self, left: IndexOpsLike, right: Any) -> SeriesOrIndex:
        from pyspark.pandas.base import column_op

        _sanitize_list_like(right)
        return column_op(Column.__gt__)(left, right)

    def astype(self, index_ops: IndexOpsLike, dtype: Union[str, type, Dtype]) -> IndexOpsLike:
        dtype, spark_type = pandas_on_spark_type(dtype)

        if isinstance(dtype, CategoricalDtype):
            return _as_categorical_type(index_ops, dtype, spark_type)
        elif isinstance(spark_type, BooleanType):
            return _as_bool_type(index_ops, dtype)
        elif isinstance(spark_type, StringType):
            return _as_string_type(index_ops, dtype)
        else:
            return _as_other_type(index_ops, dtype, spark_type)


class MapOps(DataTypeOps):
    """
    The class for binary operations of pandas-on-Spark objects with MapType.
    """

    @property
    def pretty_name(self) -> str:
        return "maps"


class StructOps(DataTypeOps):
    """
    The class for binary operations of pandas-on-Spark objects with StructType.
    """

    @property
    def pretty_name(self) -> str:
        return "structs"

    def lt(self, left: IndexOpsLike, right: Any) -> SeriesOrIndex:
        from pyspark.pandas.base import column_op

        _sanitize_list_like(right)
        return column_op(Column.__lt__)(left, right)

    def le(self, left: IndexOpsLike, right: Any) -> SeriesOrIndex:
        from pyspark.pandas.base import column_op

        _sanitize_list_like(right)
        return column_op(Column.__le__)(left, right)

    def ge(self, left: IndexOpsLike, right: Any) -> SeriesOrIndex:
        from pyspark.pandas.base import column_op

        _sanitize_list_like(right)
        return column_op(Column.__ge__)(left, right)

    def gt(self, left: IndexOpsLike, right: Any) -> SeriesOrIndex:
        from pyspark.pandas.base import column_op

        _sanitize_list_like(right)
        return column_op(Column.__gt__)(left, right)



# -*- coding: utf-8 -*-
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import inspect
import array as pyarray

import numpy as np

from pyspark import keyword_only
from pyspark.ml.classification import LogisticRegression
from pyspark.ml.clustering import KMeans
from pyspark.ml.feature import (
    Binarizer,
    Bucketizer,
    ElementwiseProduct,
    IndexToString,
    MaxAbsScaler,
    VectorSlicer,
    Word2Vec,
)
from pyspark.ml.linalg import DenseVector, SparseVector, Vectors
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.ml.param.shared import HasInputCol, HasMaxIter, HasSeed
from pyspark.ml.regression import LinearRegressionModel, GeneralizedLinearRegressionModel
from pyspark.ml.wrapper import JavaParams
from pyspark.testing.mlutils import check_params, PySparkTestCase, SparkSessionTestCase


class ParamTypeConversionTests(PySparkTestCase):
    """
    Test that param type conversion happens.
    """

    def test_int(self):
        lr = LogisticRegression(maxIter=5.0)
        self.assertEqual(lr.getMaxIter(), 5)
        self.assertTrue(isinstance(lr.getMaxIter(), int))
        self.assertRaises(TypeError, lambda: LogisticRegression(maxIter="notAnInt"))
        self.assertRaises(TypeError, lambda: LogisticRegression(maxIter=5.1))

    def test_float(self):
        lr = LogisticRegression(tol=1)
        self.assertEqual(lr.getTol(), 1.0)
        self.assertTrue(isinstance(lr.getTol(), float))
        self.assertRaises(TypeError, lambda: LogisticRegression(tol="notAFloat"))

    def test_vector(self):
        ewp = ElementwiseProduct(scalingVec=[1, 3])
        self.assertEqual(ewp.getScalingVec(), DenseVector([1.0, 3.0]))
        ewp = ElementwiseProduct(scalingVec=np.array([1.2, 3.4]))
        self.assertEqual(ewp.getScalingVec(), DenseVector([1.2, 3.4]))
        self.assertRaises(TypeError, lambda: ElementwiseProduct(scalingVec=["a", "b"]))

    def test_list(self):
        lst = [0, 1]
        for lst_like in [
            lst,
            np.array(lst),
            DenseVector(lst),
            SparseVector(len(lst), range(len(lst)), lst),
            pyarray.array("l", lst),
            range(2),
            tuple(lst),
        ]:
            converted = TypeConverters.toList(lst_like)
            self.assertIsInstance(converted, list)
            self.assertListEqual(converted, lst)

    def test_list_int(self):
        for indices in [
            [1.0, 2.0],
            np.array([1.0, 2.0]),
            DenseVector([1.0, 2.0]),
            SparseVector(2, {0: 1.0, 1: 2.0}),
            range(1, 3),
            (1.0, 2.0),
            pyarray.array("d", [1.0, 2.0]),
        ]:
            vs = VectorSlicer(indices=indices)
            self.assertListEqual(vs.getIndices(), [1, 2])
            self.assertTrue(all([isinstance(v, int) for v in vs.getIndices()]))
        self.assertRaises(TypeError, lambda: VectorSlicer(indices=["a", "b"]))

    def test_list_float(self):
        b = Bucketizer(splits=[1, 4])
        self.assertEqual(b.getSplits(), [1.0, 4.0])
        self.assertTrue(all([isinstance(v, float) for v in b.getSplits()]))
        self.assertRaises(TypeError, lambda: Bucketizer(splits=["a", 1.0]))

    def test_list_list_float(self):
        b = Bucketizer(splitsArray=[[-0.1, 0.5, 3], [-5, 1.5]])
        self.assertEqual(b.getSplitsArray(), [[-0.1, 0.5, 3.0], [-5.0, 1.5]])
        self.assertTrue(all([isinstance(v, list) for v in b.getSplitsArray()]))
        self.assertTrue(all([isinstance(v, float) for v in b.getSplitsArray()[0]]))
        self.assertTrue(all([isinstance(v, float) for v in b.getSplitsArray()[1]]))
        self.assertRaises(TypeError, lambda: Bucketizer(splitsArray=["a", 1.0]))
        self.assertRaises(TypeError, lambda: Bucketizer(splitsArray=[[-5, 1.5], ["a", 1.0]]))

    def test_list_string(self):
        for labels in [np.array(["a", "b"]), ["a", "b"], np.array(["a", "b"])]:
            idx_to_string = IndexToString(labels=labels)
            self.assertListEqual(idx_to_string.getLabels(), ["a", "b"])
        self.assertRaises(TypeError, lambda: IndexToString(labels=["a", 2]))

    def test_string(self):
        lr = LogisticRegression()
        for col in ["features", "features", np.str_("features")]:
            lr.setFeaturesCol(col)
            self.assertEqual(lr.getFeaturesCol(), "features")
        self.assertRaises(TypeError, lambda: LogisticRegression(featuresCol=2.3))

    def test_bool(self):
        self.assertRaises(TypeError, lambda: LogisticRegression(fitIntercept=1))
        self.assertRaises(TypeError, lambda: LogisticRegression(fitIntercept="false"))


class TestParams(HasMaxIter, HasInputCol, HasSeed):
    """
    A subclass of Params mixed with HasMaxIter, HasInputCol and HasSeed.
    """

    @keyword_only
    def __init__(self, seed=None):
        super().__init__()
        self._setDefault(maxIter=10)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @keyword_only
    def setParams(self, seed=None):
        """
        setParams(self, seed=None)
        Sets params for this test.
        """
        kwargs = self._input_kwargs
        return self._set(**kwargs)


class OtherTestParams(HasMaxIter, HasInputCol, HasSeed):
    """
    A subclass of Params mixed with HasMaxIter, HasInputCol and HasSeed.
    """

    @keyword_only
    def __init__(self, seed=None):
        super().__init__()
        self._setDefault(maxIter=10)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @keyword_only
    def setParams(self, seed=None):
        """
        setParams(self, seed=None)
        Sets params for this test.
        """
        kwargs = self._input_kwargs
        return self._set(**kwargs)


class HasThrowableProperty(Params):
    def __init__(self):
        super().__init__()
        self.p = Param(self, "none", "empty param")

    @property
    def test_property(self):
        raise RuntimeError("Test property to raise error when invoked")


class ParamTests(SparkSessionTestCase):
    def test_copy_new_parent(self):
        testParams = TestParams()
        # Copying an instantiated param should fail
        with self.assertRaises(ValueError):
            testParams.maxIter._copy_new_parent(testParams)
        # Copying a dummy param should succeed
        TestParams.maxIter._copy_new_parent(testParams)
        maxIter = testParams.maxIter
        self.assertEqual(maxIter.name, "maxIter")
        self.assertEqual(maxIter.doc, "max number of iterations (>= 0).")
        self.assertTrue(maxIter.parent == testParams.uid)

    def test_param(self):
        testParams = TestParams()
        maxIter = testParams.maxIter
        self.assertEqual(maxIter.name, "maxIter")
        self.assertEqual(maxIter.doc, "max number of iterations (>= 0).")
        self.assertTrue(maxIter.parent == testParams.uid)

    def test_hasparam(self):
        testParams = TestParams()
        self.assertTrue(all([testParams.hasParam(p.name) for p in testParams.params]))
        self.assertFalse(testParams.hasParam("notAParameter"))
        self.assertTrue(testParams.hasParam("maxIter"))

    def test_resolveparam(self):
        testParams = TestParams()
        self.assertEqual(testParams._resolveParam(testParams.maxIter), testParams.maxIter)
        self.assertEqual(testParams._resolveParam("maxIter"), testParams.maxIter)

        self.assertEqual(testParams._resolveParam("maxIter"), testParams.maxIter)
        self.assertRaises(AttributeError, lambda: testParams._resolveParam("아"))

        # Invalid type
        invalid_type = 1
        self.assertRaises(TypeError, testParams._resolveParam, invalid_type)

    def test_params(self):
        testParams = TestParams()
        maxIter = testParams.maxIter
        inputCol = testParams.inputCol
        seed = testParams.seed

        params = testParams.params
        self.assertEqual(params, [inputCol, maxIter, seed])

        self.assertTrue(testParams.hasParam(maxIter.name))
        self.assertTrue(testParams.hasDefault(maxIter))
        self.assertFalse(testParams.isSet(maxIter))
        self.assertTrue(testParams.isDefined(maxIter))
        self.assertEqual(testParams.getMaxIter(), 10)

        self.assertTrue(testParams.hasParam(inputCol.name))
        self.assertFalse(testParams.hasDefault(inputCol))
        self.assertFalse(testParams.isSet(inputCol))
        self.assertFalse(testParams.isDefined(inputCol))
        with self.assertRaises(KeyError):
            testParams.getInputCol()

        otherParam = Param(
            Params._dummy(),
            "otherParam",
            "Parameter used to test that " + "set raises an error for a non-member parameter.",
            typeConverter=TypeConverters.toString,
        )
        with self.assertRaises(ValueError):
            testParams.set(otherParam, "value")

        # Since the default is normally random, set it to a known number for debug str
        testParams._setDefault(seed=41)

        self.assertEqual(
            testParams.explainParams(),
            "\n".join(
                [
                    "inputCol: input column name. (undefined)",
                    "maxIter: max number of iterations (>= 0). (default: 10)",
                    "seed: random seed. (default: 41)",
                ]
            ),
        )

    def test_clear_param(self):
        df = self.spark.createDataFrame([(Vectors.dense([1.0]),), (Vectors.dense([2.0]),)], ["a"])
        maScaler = MaxAbsScaler(inputCol="a", outputCol="scaled")
        model = maScaler.fit(df)
        self.assertTrue(model.isSet(model.outputCol))
        self.assertEqual(model.getOutputCol(), "scaled")
        model.clear(model.outputCol)
        self.assertFalse(model.isSet(model.outputCol))
        self.assertEqual(model.getOutputCol()[:12], "MaxAbsScaler")
        output = model.transform(df)
        self.assertEqual(model.getOutputCol(), output.schema.names[1])

    def test_kmeans_param(self):
        algo = KMeans()
        self.assertEqual(algo.getInitMode(), "k-means||")
        algo.setK(10)
        self.assertEqual(algo.getK(), 10)
        algo.setInitSteps(10)
        self.assertEqual(algo.getInitSteps(), 10)
        self.assertEqual(algo.getDistanceMeasure(), "euclidean")
        algo.setDistanceMeasure("cosine")
        self.assertEqual(algo.getDistanceMeasure(), "cosine")

    def test_hasseed(self):
        noSeedSpecd = TestParams()
        withSeedSpecd = TestParams(seed=42)
        other = OtherTestParams()
        # Check that we no longer use 42 as the magic number
        self.assertNotEqual(noSeedSpecd.getSeed(), 42)
        origSeed = noSeedSpecd.getSeed()
        # Check that we only compute the seed once
        self.assertEqual(noSeedSpecd.getSeed(), origSeed)
        # Check that a specified seed is honored
        self.assertEqual(withSeedSpecd.getSeed(), 42)
        # Check that a different class has a different seed
        self.assertNotEqual(other.getSeed(), noSeedSpecd.getSeed())

    def test_param_property_error(self):
        param_store = HasThrowableProperty()
        self.assertRaises(RuntimeError, lambda: param_store.test_property)
        params = param_store.params  # should not invoke the property 'test_property'
        self.assertEqual(len(params), 1)

    def test_word2vec_param(self):
        model = Word2Vec().setWindowSize(6)
        # Check windowSize is set properly
        self.assertEqual(model.getWindowSize(), 6)

    def test_copy_param_extras(self):
        tp = TestParams(seed=42)
        extra = {tp.getParam(TestParams.inputCol.name): "copy_input"}
        tp_copy = tp.copy(extra=extra)
        self.assertEqual(tp.uid, tp_copy.uid)
        self.assertEqual(tp.params, tp_copy.params)
        for k, v in extra.items():
            self.assertTrue(tp_copy.isDefined(k))
            self.assertEqual(tp_copy.getOrDefault(k), v)
        copied_no_extra = {}
        for k, v in tp_copy._paramMap.items():
            if k not in extra:
                copied_no_extra[k] = v
        self.assertEqual(tp._paramMap, copied_no_extra)
        self.assertEqual(tp._defaultParamMap, tp_copy._defaultParamMap)
        with self.assertRaises(TypeError):
            tp.copy(extra={"unknown_parameter": None})
        with self.assertRaises(TypeError):
            tp.copy(extra=["must be a dict"])

    def test_logistic_regression_check_thresholds(self):
        self.assertIsInstance(
            LogisticRegression(threshold=0.5, thresholds=[0.5, 0.5]), LogisticRegression
        )

        self.assertRaisesRegex(
            ValueError,
            "Logistic Regression getThreshold found inconsistent.*$",
            LogisticRegression,
            threshold=0.42,
            thresholds=[0.5, 0.5],
        )

    def test_preserve_set_state(self):
        dataset = self.spark.createDataFrame([(0.5,)], ["data"])
        binarizer = Binarizer(inputCol="data")
        self.assertFalse(binarizer.isSet("threshold"))
        binarizer.transform(dataset)
        binarizer._transfer_params_from_java()
        self.assertFalse(
            binarizer.isSet("threshold"),
            "Params not explicitly set should remain unset after transform",
        )

    def test_default_params_transferred(self):
        dataset = self.spark.createDataFrame([(0.5,)], ["data"])
        binarizer = Binarizer(inputCol="data")
        # intentionally change the pyspark default, but don't set it
        binarizer._defaultParamMap[binarizer.outputCol] = "my_default"
        result = binarizer.transform(dataset).select("my_default").collect()
        self.assertFalse(binarizer.isSet(binarizer.outputCol))
        self.assertEqual(result[0][0], 1.0)

    def test_lr_evaluate_invalid_type(self):
        lr = LinearRegressionModel()
        invalid_type = ""
        self.assertRaises(TypeError, lr.evaluate, invalid_type)

    def test_glr_evaluate_invalid_type(self):
        glr = GeneralizedLinearRegressionModel()
        invalid_type = ""
        self.assertRaises(TypeError, glr.evaluate, invalid_type)


class DefaultValuesTests(PySparkTestCase):
    """
    Test :py:class:`JavaParams` classes to see if their default Param values match
    those in their Scala counterparts.
    """

    def test_java_params(self):
        import re

        import pyspark.ml.feature
        import pyspark.ml.classification
        import pyspark.ml.clustering
        import pyspark.ml.evaluation
        import pyspark.ml.pipeline
        import pyspark.ml.recommendation
        import pyspark.ml.regression

        modules = [
            pyspark.ml.feature,
            pyspark.ml.classification,
            pyspark.ml.clustering,
            pyspark.ml.evaluation,
            pyspark.ml.pipeline,
            pyspark.ml.recommendation,
            pyspark.ml.regression,
        ]
        for module in modules:
            for name, cls in inspect.getmembers(module, inspect.isclass):
                if (
                    not name.endswith("Model")
                    and not name.endswith("Params")
                    and issubclass(cls, JavaParams)
                    and not inspect.isabstract(cls)
                    and not re.match("_?Java", name)
                    and name != "_LSH"
                    and name != "_Selector"
                ):
                    check_params(self, cls(), check_params_exist=True)

        # Additional classes that need explicit construction
        from pyspark.ml.feature import CountVectorizerModel, StringIndexerModel

        check_params(
            self, CountVectorizerModel.from_vocabulary(["a"], "input"), check_params_exist=True
        )
        check_params(
            self, StringIndexerModel.from_labels(["a", "b"], "input"), check_params_exist=True
        )


if __name__ == "__main__":
    from pyspark.testing import main

    main()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import tempfile

import numpy as np

from pyspark.ml.linalg import Vectors, SparseVector
from pyspark.ml.clustering import (
    KMeans,
    KMeansModel,
    KMeansSummary,
    BisectingKMeans,
    BisectingKMeansModel,
    BisectingKMeansSummary,
    GaussianMixture,
    GaussianMixtureModel,
    GaussianMixtureSummary,
    LDA,
    LDAModel,
    LocalLDAModel,
    DistributedLDAModel,
    PowerIterationClustering,
)
from pyspark.testing.sqlutils import ReusedSQLTestCase


class ClusteringTestsMixin:
    def test_kmeans(self):
        df = (
            self.spark.createDataFrame(
                [
                    (1, 1.0, Vectors.dense([-0.1, -0.05])),
                    (2, 2.0, Vectors.dense([-0.01, -0.1])),
                    (3, 3.0, Vectors.dense([0.9, 0.8])),
                    (4, 1.0, Vectors.dense([0.75, 0.935])),
                    (5, 1.0, Vectors.dense([-0.83, -0.68])),
                    (6, 1.0, Vectors.dense([-0.91, -0.76])),
                ],
                ["index", "weight", "features"],
            )
            .coalesce(1)
            .sortWithinPartitions("index")
            .select("weight", "features")
        )

        km = KMeans(
            k=2,
            maxIter=2,
            weightCol="weight",
        )
        self.assertEqual(km.getK(), 2)
        self.assertEqual(km.getMaxIter(), 2)
        self.assertEqual(km.getWeightCol(), "weight")

        model = km.fit(df)
        self.assertEqual(km.uid, model.uid)

        centers = model.clusterCenters()
        self.assertEqual(len(centers), 2)
        self.assertEqual(model.numFeatures, 2)
        self.assertTrue(np.allclose(centers[0], [-0.372, -0.338], atol=1e-3), centers[0])
        self.assertTrue(np.allclose(centers[1], [0.8625, 0.83375], atol=1e-3), centers[1])

        output = model.transform(df)
        expected_cols = ["weight", "features", "prediction"]
        self.assertEqual(output.columns, expected_cols)
        self.assertEqual(output.count(), 6)

        self.assertTrue(np.allclose(model.predict(Vectors.dense(0.0, 5.0)), 1, atol=1e-4))

        def check_summary():
            # Model summary
            self.assertTrue(model.hasSummary)
            summary = model.summary
            self.assertTrue(isinstance(summary, KMeansSummary))
            self.assertEqual(summary.k, 2)
            self.assertEqual(summary.numIter, 2)
            self.assertEqual(summary.clusterSizes, [4, 2])
            self.assertTrue(np.allclose(summary.trainingCost, 1.35710375, atol=1e-4))

            self.assertEqual(summary.featuresCol, "features")
            self.assertEqual(summary.predictionCol, "prediction")

            self.assertEqual(summary.cluster.columns, ["prediction"])
            self.assertEqual(summary.cluster.count(), 6)

            self.assertEqual(summary.predictions.columns, expected_cols)
            self.assertEqual(summary.predictions.count(), 6)

        # check summary before model offloading occurs
        check_summary()

        # save & load
        with tempfile.TemporaryDirectory(prefix="kmeans_model") as d:
            km.write().overwrite().save(d)
            km2 = KMeans.load(d)
            self.assertEqual(str(km), str(km2))

            model.write().overwrite().save(d)
            model2 = KMeansModel.load(d)
            self.assertEqual(str(model), str(model2))
            self.assertFalse(model2.hasSummary)
            with self.assertRaisesRegex(Exception, "No training summary available"):
                model2.summary

    def test_bisecting_kmeans(self):
        df = (
            self.spark.createDataFrame(
                [
                    (1, 1.0, Vectors.dense([-0.1, -0.05])),
                    (2, 2.0, Vectors.dense([-0.01, -0.1])),
                    (3, 3.0, Vectors.dense([0.9, 0.8])),
                    (4, 1.0, Vectors.dense([0.75, 0.935])),
                    (5, 1.0, Vectors.dense([-0.83, -0.68])),
                    (6, 1.0, Vectors.dense([-0.91, -0.76])),
                ],
                ["index", "weight", "features"],
            )
            .coalesce(1)
            .sortWithinPartitions("index")
            .select("weight", "features")
        )

        bkm = BisectingKMeans(
            k=2,
            maxIter=2,
            minDivisibleClusterSize=1.0,
            weightCol="weight",
        )
        self.assertEqual(bkm.getK(), 2)
        self.assertEqual(bkm.getMaxIter(), 2)
        self.assertEqual(bkm.getMinDivisibleClusterSize(), 1.0)
        self.assertEqual(bkm.getWeightCol(), "weight")

        model = bkm.fit(df)
        self.assertEqual(bkm.uid, model.uid)

        centers = model.clusterCenters()
        self.assertEqual(len(centers), 2)
        self.assertEqual(model.numFeatures, 2)
        self.assertTrue(np.allclose(centers[0], [-0.372, -0.338], atol=1e-3), centers[0])
        self.assertTrue(np.allclose(centers[1], [0.8625, 0.83375], atol=1e-3), centers[1])

        output = model.transform(df)
        expected_cols = ["weight", "features", "prediction"]
        self.assertEqual(output.columns, expected_cols)
        self.assertEqual(output.count(), 6)

        self.assertTrue(np.allclose(model.predict(Vectors.dense(0.0, 5.0)), 1, atol=1e-4))

        # BisectingKMeans-specific method: computeCost
        self.assertTrue(np.allclose(model.computeCost(df), 1.164325125, atol=1e-4))

        # Model summary
        self.assertTrue(model.hasSummary)
        summary = model.summary
        self.assertTrue(isinstance(summary, BisectingKMeansSummary))
        self.assertEqual(summary.k, 2)
        self.assertEqual(summary.numIter, 2)
        self.assertEqual(summary.clusterSizes, [4, 2])
        self.assertTrue(np.allclose(summary.trainingCost, 1.3571037499999998, atol=1e-4))

        self.assertEqual(summary.featuresCol, "features")
        self.assertEqual(summary.predictionCol, "prediction")

        self.assertEqual(summary.cluster.columns, ["prediction"])
        self.assertEqual(summary.cluster.count(), 6)

        self.assertEqual(summary.predictions.columns, expected_cols)
        self.assertEqual(summary.predictions.count(), 6)

        # save & load
        with tempfile.TemporaryDirectory(prefix="bisecting_kmeans") as d:
            bkm.write().overwrite().save(d)
            bkm2 = BisectingKMeans.load(d)
            self.assertEqual(str(bkm), str(bkm2))

            model.write().overwrite().save(d)
            model2 = BisectingKMeansModel.load(d)
            self.assertEqual(str(model), str(model2))

    def test_gaussian_mixture(self):
        spark = self.spark
        df = (
            spark.createDataFrame(
                [
                    (1, 1.0, Vectors.dense([-0.1, -0.05])),
                    (2, 2.0, Vectors.dense([-0.01, -0.1])),
                    (3, 3.0, Vectors.dense([0.9, 0.8])),
                    (4, 1.0, Vectors.dense([0.75, 0.935])),
                    (5, 1.0, Vectors.dense([-0.83, -0.68])),
                    (6, 1.0, Vectors.dense([-0.91, -0.76])),
                ],
                ["index", "weight", "features"],
            )
            .coalesce(1)
            .sortWithinPartitions("index")
            .select("weight", "features")
        )

        gmm = GaussianMixture(
            k=2,
            maxIter=2,
            weightCol="weight",
            seed=1,
        )
        self.assertEqual(gmm.getK(), 2)
        self.assertEqual(gmm.getMaxIter(), 2)
        self.assertEqual(gmm.getWeightCol(), "weight")
        self.assertEqual(gmm.getSeed(), 1)

        model = gmm.fit(df)
        self.assertEqual(gmm.uid, model.uid)
        self.assertEqual(model.numFeatures, 2)
        self.assertEqual(len(model.weights), 2)
        self.assertTrue(
            np.allclose(model.weights, [0.541014115744985, 0.4589858842550149], atol=1e-4),
            model.weights,
        )

        # check the gaussians
        gaussians = model.gaussians
        self.assertEqual(len(gaussians), 2)
        self.assertTrue(
            np.allclose(
                gaussians[0].mean.toArray(),
                [0.28586084899633746, 0.28513455726904297],
                atol=1e-4,
            ),
            gaussians[0].mean,
        )
        self.assertTrue(
            np.allclose(
                gaussians[0].cov.toArray(),
                [[0.41732752, 0.38378601], [0.38378601, 0.36454957]],
                atol=1e-4,
            ),
            gaussians[0].cov,
        )
        self.assertTrue(
            np.allclose(
                gaussians[1].mean.toArray(),
                [0.04795771063097124, 0.06212817950777127],
                atol=1e-4,
            ),
            gaussians[1].mean,
        )
        self.assertTrue(
            np.allclose(
                gaussians[1].cov.toArray(),
                [[0.50359595, 0.44696663], [0.44696663, 0.40424231]],
                atol=1e-4,
            ),
            gaussians[1].cov,
        )

        self.assertEqual(model.gaussiansDF.columns, ["mean", "cov"])
        self.assertEqual(model.gaussiansDF.count(), 2)

        vec = Vectors.dense(0.0, 5.0)
        pred = model.predict(vec)
        self.assertTrue(np.allclose(pred, 0, atol=1e-4), pred)
        pred = model.predictProbability(vec)
        self.assertTrue(np.allclose(pred.toArray(), [0.5, 0.5], atol=1e-4), pred)

        output = model.transform(df)
        expected_cols = ["weight", "features", "probability", "prediction"]
        self.assertEqual(output.columns, expected_cols)
        self.assertEqual(output.count(), 6)

        def check_summary():
            # Model summary
            self.assertTrue(model.hasSummary)
            summary = model.summary
            self.assertTrue(isinstance(summary, GaussianMixtureSummary))
            self.assertEqual(summary.k, 2)
            self.assertEqual(summary.numIter, 2)
            self.assertEqual(len(summary.clusterSizes), 2)
            self.assertEqual(summary.clusterSizes, [3, 3])
            ll = summary.logLikelihood
            self.assertTrue(ll < 0, ll)
            self.assertTrue(np.allclose(ll, -1.311264553744033, atol=1e-4), ll)

            self.assertEqual(summary.featuresCol, "features")
            self.assertEqual(summary.predictionCol, "prediction")
            self.assertEqual(summary.probabilityCol, "probability")

            self.assertEqual(summary.cluster.columns, ["prediction"])
            self.assertEqual(summary.cluster.count(), 6)

            self.assertEqual(summary.predictions.columns, expected_cols)
            self.assertEqual(summary.predictions.count(), 6)

            self.assertEqual(summary.probability.columns, ["probability"])
            self.assertEqual(summary.predictions.count(), 6)

        # save & load
        with tempfile.TemporaryDirectory(prefix="gaussian_mixture") as d:
            gmm.write().overwrite().save(d)
            gmm2 = GaussianMixture.load(d)
            self.assertEqual(str(gmm), str(gmm2))

            model.write().overwrite().save(d)
            model2 = GaussianMixtureModel.load(d)
            self.assertEqual(str(model), str(model2))

    def test_local_lda(self):
        spark = self.spark
        df = (
            spark.createDataFrame(
                [
                    [1, Vectors.dense([0.0, 1.0])],
                    [2, SparseVector(2, {0: 1.0})],
                ],
                ["id", "features"],
            )
            .coalesce(1)
            .sortWithinPartitions("id")
        )

        lda = LDA(k=2, optimizer="online", seed=1)
        lda.setMaxIter(1)
        self.assertEqual(lda.getK(), 2)
        self.assertEqual(lda.getOptimizer(), "online")
        self.assertEqual(lda.getMaxIter(), 1)
        self.assertEqual(lda.getSeed(), 1)

        model = lda.fit(df)
        self.assertEqual(lda.uid, model.uid)
        self.assertIsInstance(model, LDAModel)
        self.assertIsInstance(model, LocalLDAModel)
        self.assertNotIsInstance(model, DistributedLDAModel)
        self.assertFalse(model.isDistributed())

        dc = model.estimatedDocConcentration()
        self.assertTrue(np.allclose(dc.toArray(), [0.5, 0.5], atol=1e-4), dc)
        topics = model.topicsMatrix()
        self.assertTrue(
            np.allclose(
                topics.toArray(), [[1.20296728, 1.15740442], [0.99357675, 1.02993164]], atol=1e-4
            ),
            topics,
        )

        ll = model.logLikelihood(df)
        self.assertTrue(np.allclose(ll, -3.2125122434040088, atol=1e-4), ll)
        lp = model.logPerplexity(df)
        self.assertTrue(np.allclose(lp, 1.6062561217020044, atol=1e-4), lp)
        dt = model.describeTopics()
        self.assertEqual(dt.columns, ["topic", "termIndices", "termWeights"])
        self.assertEqual(dt.count(), 2)

        # LocalLDAModel specific methods
        self.assertEqual(model.vocabSize(), 2)

        output = model.transform(df)
        expected_cols = ["id", "features", "topicDistribution"]
        self.assertEqual(output.columns, expected_cols)
        self.assertEqual(output.count(), 2)

        # save & load
        with tempfile.TemporaryDirectory(prefix="local_lda") as d:
            lda.write().overwrite().save(d)
            lda2 = LDA.load(d)
            self.assertEqual(str(lda), str(lda2))

            model.write().overwrite().save(d)
            model2 = LocalLDAModel.load(d)
            self.assertEqual(str(model), str(model2))

    def test_distributed_lda(self):
        spark = self.spark
        df = (
            spark.createDataFrame(
                [
                    [1, Vectors.dense([0.0, 1.0])],
                    [2, SparseVector(2, {0: 1.0})],
                ],
                ["id", "features"],
            )
            .coalesce(1)
            .sortWithinPartitions("id")
        )

        lda = LDA(k=2, optimizer="em", seed=1)
        lda.setMaxIter(1)

        self.assertEqual(lda.getK(), 2)
        self.assertEqual(lda.getOptimizer(), "em")
        self.assertEqual(lda.getMaxIter(), 1)
        self.assertEqual(lda.getSeed(), 1)

        model = lda.fit(df)
        self.assertEqual(lda.uid, model.uid)
        self.assertIsInstance(model, LDAModel)
        self.assertNotIsInstance(model, LocalLDAModel)
        self.assertIsInstance(model, DistributedLDAModel)
        self.assertTrue(model.isDistributed())
        self.assertEqual(model.vocabSize(), 2)

        dc = model.estimatedDocConcentration()
        self.assertTrue(np.allclose(dc.toArray(), [26.0, 26.0], atol=1e-4), dc)
        topics = model.topicsMatrix()
        self.assertTrue(
            np.allclose(
                topics.toArray(), [[0.39149926, 0.60850074], [0.60991237, 0.39008763]], atol=1e-4
            ),
            topics,
        )

        ll = model.logLikelihood(df)
        self.assertTrue(np.allclose(ll, -3.719138517085772, atol=1e-4), ll)
        lp = model.logPerplexity(df)
        self.assertTrue(np.allclose(lp, 1.859569258542886, atol=1e-4), lp)

        dt = model.describeTopics()
        self.assertEqual(dt.columns, ["topic", "termIndices", "termWeights"])
        self.assertEqual(dt.count(), 2)

        # DistributedLDAModel specific methods
        ll = model.trainingLogLikelihood()
        self.assertTrue(np.allclose(ll, -1.3847360462201639, atol=1e-4), ll)
        lp = model.logPrior()
        self.assertTrue(np.allclose(lp, -69.59963186898915, atol=1e-4), lp)
        model.getCheckpointFiles()

        output = model.transform(df)
        expected_cols = ["id", "features", "topicDistribution"]
        self.assertEqual(output.columns, expected_cols)
        self.assertEqual(output.count(), 2)

        # Test toLocal()
        localModel = model.toLocal()
        self.assertIsInstance(localModel, LDAModel)
        self.assertIsInstance(localModel, LocalLDAModel)
        self.assertNotIsInstance(localModel, DistributedLDAModel)
        self.assertFalse(localModel.isDistributed())
        output = localModel.transform(df)
        expected_cols = ["id", "features", "topicDistribution"]
        self.assertEqual(output.columns, expected_cols)
        self.assertEqual(output.count(), 2)

        # save & load
        with tempfile.TemporaryDirectory(prefix="distributed_lda") as d:
            lda.write().overwrite().save(d)
            lda2 = LDA.load(d)
            self.assertEqual(str(lda), str(lda2))

            model.write().overwrite().save(d)
            model2 = DistributedLDAModel.load(d)
            self.assertEqual(str(model), str(model2))

    def test_power_iteration_clustering(self):
        spark = self.spark

        data = [
            (1, 0, 0.5),
            (2, 0, 0.5),
            (2, 1, 0.7),
            (3, 0, 0.5),
            (3, 1, 0.7),
            (3, 2, 0.9),
            (4, 0, 0.5),
            (4, 1, 0.7),
            (4, 2, 0.9),
            (4, 3, 1.1),
            (5, 0, 0.5),
            (5, 1, 0.7),
            (5, 2, 0.9),
            (5, 3, 1.1),
            (5, 4, 1.3),
        ]
        df = spark.createDataFrame(data, ["src", "dst", "weight"]).repartition(1)

        pic = PowerIterationClustering(k=2, weightCol="weight")
        pic.setMaxIter(40)

        self.assertEqual(pic.getK(), 2)
        self.assertEqual(pic.getMaxIter(), 40)
        self.assertEqual(pic.getWeightCol(), "weight")

        assignments = pic.assignClusters(df)
        self.assertEqual(assignments.columns, ["id", "cluster"])
        self.assertEqual(assignments.count(), 6)

        # save & load
        with tempfile.TemporaryDirectory(prefix="power_iteration_clustering") as d:
            pic.write().overwrite().save(d)
            pic2 = PowerIterationClustering.load(d)
            self.assertEqual(str(pic), str(pic2))
            self.assertEqual(pic.uid, pic2.uid)
            self.assertEqual(pic.getK(), pic2.getK())
            self.assertEqual(pic.getMaxIter(), pic2.getMaxIter())
            self.assertEqual(pic.getWeightCol(), pic2.getWeightCol())


class ClusteringTests(ClusteringTestsMixin, ReusedSQLTestCase):
    pass


if __name__ == "__main__":
    from pyspark.testing import main

    main()
